#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2020 Andrei Leonard Nicusan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : adac_forte.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 01.04.2021


import  os
import  time
import  textwrap
from    glob            import  glob

import  numpy           as      np
from    natsort         import  natsorted
from    pept            import  LineData

from    .extensions     import  convert_adac_forte




def adac_forte(
    filepath,
    sample_size = None,
    overlap = None,
    verbose = True,
):
    '''Initialise PEPT lines of response (LoRs) from a binary file outputted by
    the ADAC Forte parallel screen detector list mode (common file extension
    ".da01").

    Parameters
    ----------
    filepath : str
        The path to a ADAC Forte-generated binary file from which the LoRs
        will be read into the `LineData` format. If you have multiple files,
        use a wildcard (*) after their common substring to concatenate them,
        e.g. "DS1.da*" will add ["DS1.da01", "DS1.da02", "DS1.da02_02"].

    sample_size : int, default 0
        An `int` that defines the number of lines that should be returned
        when iterating over `lines`. A `sample_size` of 0 yields all the
        data as one single sample. A good starting value would be 200 times
        the maximum number of tracers that would be tracked.

    overlap : int, default 0
        An `int` that defines the overlap between two consecutive samples
        that are returned when iterating over `lines`. An overlap of 0
        implies consecutive samples, while an overlap of
        (`sample_size` - 1) means incrementing the samples by one. A
        negative overlap means skipping values between samples. An error is
        raised if `overlap` is larger than or equal to `sample_size`.

    verbose : bool, default True
        An option that enables printing the time taken for the
        initialisation of an instance of the class. Useful when reading
        large files (10gb files for PEPT data is not unheard of).

    Returns
    -------
    LineData
        The initialised LoRs.

    Raises
    ------
    FileNotFoundError
        If the input `filepath` does not exist.

    ValueError
        If `overlap` >= `sample_size`. Overlap has to be smaller than
        `sample_size`. Note that it can also be negative.

    Examples
    --------
    Initialise a `ParallelScreens` array for three LoRs on a parallel screens
    PEPT scanner (i.e. each line is defined by **two** points each) with a
    head separation of 500 mm:

    >>> lors = pept.scanners.adac_forte("binary_data_adac.da01")
    Initialised the PEPT data in 0.011 s.

    >>> lors
    LineData
    --------
    sample_size = 0
    overlap =     0
    samples =     1
    lines =
      [[0.00000000e+00 1.62250000e+02 3.60490000e+02 ... 4.14770000e+02
        3.77010000e+02 3.10000000e+02]
       [4.19512195e-01 2.05910000e+02 2.68450000e+02 ... 3.51640000e+02
        2.95000000e+02 3.10000000e+02]
       [8.39024390e-01 3.16830000e+02 1.26260000e+02 ... 2.74350000e+02
        3.95300000e+02 3.10000000e+02]
       ...
       [1.98255892e+04 2.64320000e+02 2.43080000e+02 ... 2.25970000e+02
        4.01200000e+02 3.10000000e+02]
       [1.98263928e+04 3.19780000e+02 3.38660000e+02 ... 2.75530000e+02
        5.19200000e+02 3.10000000e+02]
       [1.98271964e+04 2.41310000e+02 4.15360000e+02 ... 2.91460000e+02
        4.63150000e+02 3.10000000e+02]]
    lines.shape = (32526, 7)
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    if verbose:
        start = time.time()

    # If we have a wildcard (*) in the filepath, find all files
    if "*" in filepath:
        filepaths = natsorted(glob(filepath))
        if verbose:
            print(f"Concatenating files:\n  {filepaths}")

    # Otherwise make sure the single file exists
    else:
        if not os.path.isfile(filepath):
            raise FileNotFoundError(textwrap.fill((
                f"The input file path {filepath} does not exist!"
            )))

        filepaths = [filepath]

    lines = convert_adac_forte(filepaths[0])

    # If there are multiple files, concatenate them (and add up the timestamps)
    for i in range(1, len(filepaths)):
        new_lines = convert_adac_forte(filepaths[i])
        new_lines[:, 0] += lines[-1, 0]

        lines = np.vstack((lines, new_lines))

    # Flip Y axis
    lines[:, [2, 5]] = 600 - lines[:, [2, 5]]

    if verbose:
        end = time.time()
        print(f"\nInitialised PEPT data in {end - start:3.3f} s.\n")

    return LineData(lines, sample_size = sample_size, overlap = overlap)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : pipelines.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 07.08.2021


import  os
import  sys
import  time
import  textwrap
from    abc                 import  ABC, abstractmethod

if sys.version_info.minor >= 9:
    # Python 3.9
    from collections.abc    import  Iterable
else:
    from typing             import  Iterable

import  numpy               as      np

from    beartype            import  beartype
from    tqdm                import  tqdm
from    joblib              import  Parallel, delayed

from    .iterable_samples   import  PEPTObject
from    .point_data         import  PointData
from    .line_data          import  LineData
from    .voxel_data         import  Voxels




class Transformer(ABC, PEPTObject):
    '''Base class for PEPT filters (transforming a sample into another) and
    reducers (transforming a list of samples).

    You should only need to subclass `Filter` and `Reducer` (or even, better,
    their more specialised subclasses, e.g. `LineDataFilter`).
    '''

    def __add__(self, other):
        # Allow adding (+) multiple filters / reducers into a pipeline
        return Pipeline([self, other])


    def __repr__(self):
        # Select all attributes from dir(obj) that are not callable (i.e. not
        # methods) and don't start with "_" (i.e. they're private)
        docs = []
        for attr in dir(self):
            memb = getattr(self, attr)
            if not callable(memb) and not attr.startswith("_"):
                # If it's a nested collection, print it on a new, indented line
                if (isinstance(memb, np.ndarray) and memb.ndim > 1) or \
                        isinstance(memb, PEPTObject):
                    memb_str = textwrap.indent(str(memb), '  ')
                    docs.append(f"{attr} = \n{memb_str}")
                else:
                    docs.append(f"{attr} = {memb}")

        return f"{type(self).__name__}(" + ", ".join(docs) + ")"




class Filter(Transformer):
    '''Abstract class from which PEPT filters inherit. You only need to define
    a method `def fit_sample(self, sample)`, which processes a *single* sample.

    If you define a filter on `LineData`, you should subclass `LineDataFilter`.
    Same goes for `PointData` with `PointDataFilter`.
    '''

    @abstractmethod
    def fit_sample(self, sample):
        pass


    @beartype
    def fit(
        self,
        samples: Iterable,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        '''Apply self.fit_sample (implemented by subclasses) according to the
        execution policy. Simply return a list of processed samples. If you
        need a reduction step (e.g. stack all processed samples), apply it
        in the subclass.
        '''

        if executor == "sequential":
            if verbose:
                samples = tqdm(samples, desc = f"{type(self).__name__} :")

            return [self.fit_sample(s) for s in samples]

        elif executor == "joblib":
            if verbose:
                samples = tqdm(samples, desc = f"{type(self).__name__} :")

            if max_workers is None:
                max_workers = os.cpu_count()

            return Parallel(n_jobs = max_workers)(
                delayed(self.fit_sample)(s) for s in samples
            )

        else:
            with executor(max_workers = max_workers) as exe:
                futures = [exe.submit(self.fit_sample, s) for s in samples]

                if verbose:
                    futures = tqdm(futures, desc = f"{type(self).__name__} :")

                return [f.result() for f in futures]




class Reducer(Transformer):
    '''Abstract class from which PEPT reducers inherit. You only need to define
    a method `def fit(self, samples)`, which processes an *iterable* of samples
    (most commonly a `LineData` or `PointData`).
    '''

    @abstractmethod
    def fit(self, samples):
        pass




class LineDataFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.LineData`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    @beartype
    def fit(
        self,
        line_data: Iterable[LineData],
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        return Filter.fit(
            self, line_data, executor, max_workers, verbose = verbose
        )




class PointDataFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.PointData`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    @beartype
    def fit(
        self,
        point_data: Iterable[PointData],
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        return Filter.fit(
            self, point_data, executor, max_workers, verbose = verbose
        )




class VoxelsFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.Voxels`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    @beartype
    def fit(
        self,
        line_data: Iterable[Voxels],
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        return Filter.fit(
            self, line_data, executor, max_workers, verbose = verbose
        )




class Pipeline(PEPTObject):
    '''A PEPT processing pipeline, chaining multiple `Filter` and `Reducer`
    for efficient, parallel execution.

    After a pipeline is constructed, the `fit(samples)` method can be called,
    which will apply the chain of filters and reducers on the samples of data.

    A filter is simply a transformation applied to a sample (e.g. `Voxelliser`
    on a single sample of `LineData`). A reducer is a transformation applied to
    a list of *all* samples (e.g. `Stack` on all samples of `PointData`).

    Note that only filters can be applied in parallel, but the great advantage
    of a `Pipeline` is that it significantly reduces the amount of data copying
    and intermediate results' storage. Reducers will require collecting all
    results.

    There are three execution policies at the moment: "sequential" is
    single-threaded (slower, but easy to debug), "joblib" (very fast on medium
    datasets due to joblib's caching) and any `concurrent.futures.Executor`
    subclass (e.g. MPIPoolExecutor for parallel processing on distributed
    clusters).

    Attributes
    ----------
    transformers : list[pept.base.Filter or pept.base.Reducer]
        A list of transformers that will be applied consecutively to samples of
        data. A `pept.base.Filter` subclass transforms individual samples of
        data (and defines a `fit_sample` method), while a `pept.base.Reducer`
        subclass transforms entire lists of samples (and defines a `fit`
        method).

    Examples
    --------
    A pipeline can be created in two ways: either by adding (+) multiple
    transformers together, or explicitly constructing the `Pipeline` class.

    The first method is the most straightforward:

    >>> import pept

    >>> filter1 = pept.tracking.Cutpoints(max_distance = 0.5)
    >>> filter2 = pept.tracking.HDBSCAN(true_fraction = 0.1)
    >>> reducer = pept.tracking.Stack()
    >>> pipeline = filter1 + filter2 + reducer

    >>> print(pipeline)
    Pipeline
    --------
    transformers = [
        Cutpoints(append_indices = False, cutoffs = None, max_distance = 0.5)
        HDBSCAN(clusterer = HDBSCAN(), max_tracers = 1, true_fraction = 0.1)
        Stack(overlap = None, sample_size = None)
    ]

    >>> lors = pept.LineData(...)        # Some samples of lines
    >>> points = pipeline.fit(lors)

    The chain of filters can also be applied to a single sample:

    >>> point = pipeline.fit_sample(lors[0])

    The pipeline's `fit` method allows specifying an execution policy:

    >>> points = pipeline.fit(lors, executor = "sequential")
    >>> points = pipeline.fit(lors, executor = "joblib")

    >>> from mpi4py.futures import MPIPoolExecutor
    >>> points = pipeline.fit(lors, executor = MPIPoolExecutor)

    The `pept.Pipeline` constructor can also be called directly, which allows
    the enumeration of filters:

    >>> pipeline = pept.Pipeline([filter1, filter2, reducer])

    Adding new filters is very easy:

    >>> pipeline_extra = pipeline + filter2

    '''

    def __init__(self, transformers):
        '''Construct the class from an iterable of ``Filter``, ``Reducer``
        and/or other ``Pipeline`` instances (which will be flattened).
        '''

        # Type-checking inputs
        transformers = list(transformers)

        # The list of transformers might contain entire pipelines; extract
        # their filters into a flat list of transformers
        flattened = []
        for i, t in enumerate(transformers):
            if isinstance(t, Transformer):
                flattened.append(t)

            elif isinstance(t, Pipeline):
                flattened.extend(t.transformers)

            else:
                raise TypeError(textwrap.fill((
                    "All input `transformers` must be subclasses of "
                    "`pept.base.Filter`, `pept.base.Reducer` or "
                    f"`pept.base.Pipeline`. Received `{type(t)}` at index {i}."
                )))

        # Setting class attributes
        self._transformers = flattened


    @property
    def filters(self):
        '''Only the `Filter` instances from the `transformers`. They can be
        applied in parallel.
        '''
        return [t for t in self.transformers if isinstance(t, Filter)]


    @property
    def reducers(self):
        '''Only the `Reducer` instances from the `transformers`. They require
        collecting all parallel results.
        '''
        return [t for t in self.transformers if isinstance(t, Reducer)]


    @property
    def transformers(self):
        '''The list of `Transformer` to be applied; this includes both `Filter`
        and `Reducer` instances.
        '''
        return self._transformers


    def _fit_sample(self, transformers, sample):
        # It is more efficient to fit samples procedurally; recursion in Python
        # is slow and doesn't allow garbage-collecting intermediate states
        for t in transformers:
            # If we need to apply a reducer to a single sample, use [sample]
            if isinstance(t, Reducer):
                sample = t.fit([sample])
            else:
                sample = t.fit_sample(sample)

        return sample


    def fit_sample(self, sample):
        '''Apply all transformers - consecutively - to a single sample of data.
        The output type is simply what the transformers return.
        '''
        return self._fit_sample(self.transformers, sample)


    def _fit(self, transformers, samples, executor, max_workers, desc = None):
        # Select type of execution of the filtering steps and extract a list
        # of processed samples
        if executor == "sequential":
            if desc is not None:
                samples = tqdm(samples, desc = desc)

            return [self._fit_sample(transformers, s) for s in samples]

        elif executor == "joblib":
            if desc is not None:
                samples = tqdm(samples, desc = desc)

            # Joblib's `max_workers` behaviour is different than for Executors.
            # The latter simply sets the maximum number of threads available,
            # while the former uses a max_worker=1. Make behaviour consistent.
            if max_workers is None:
                max_workers = os.cpu_count()

            return Parallel(n_jobs = max_workers)(
                delayed(self._fit_sample)(transformers, s) for s in samples
            )

        else:
            # Otherwise assume `executor` is a `concurrent.futures.Executor`
            # subclass (e.g. ProcessPoolExecutor, MPIPoolExecutor).
            with executor(max_workers = max_workers) as exe:
                futures = [
                    exe.submit(self._fit_sample, transformers, s)
                    for s in samples
                ]

                if desc is not None:
                    futures = tqdm(futures, desc = desc)

                return [f.result() for f in futures]


    @beartype
    def fit(
        self,
        samples: Iterable,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        '''Apply all transformers defined to all `samples`. Filters are applied
        according to the `executor` policy (e.g. parallel via "joblib"), while
        reducers are applied on a single thread.

        Parameters
        ----------
        samples : IterableSamples
            Any subclass of `IterableSamples` (e.g. `pept.LineData`) that
            allows iterating through samples of data.

        executor : "sequential", "joblib", or `concurrent.futures.Executor` \
                subclass, default "joblib"
            The execution policy controlling how the chain of filters are
            applied to each sample in `samples`; "sequential" is single
            threaded (slow, but easy to debug), "joblib" is multi-threaded
            (very fast due to joblib's caching). Alternatively, a
            `concurrent.futures.Executor` subclass can be used (e.g.
            `MPIPoolExecutor` for distributed computing on clusters).

        max_workers : int, optional
            The maximum number of workers to use for parallel executors. If
            `None` (default), the maximum number of CPUs are used.

        verbose : bool, default True
            If True, show extra information during processing, e.g. loading
            bars.
        '''

        # If verbose, time operation
        if verbose:
            start = time.time()

        # Aggregate processing steps into a list where consecutive filters are
        # collapsed into tuples
        steps = self.steps()
        if verbose:
            nbatches = sum((isinstance(step, tuple) for step in steps))

        nbatch = 0
        for step in steps:
            # Apply continuous sequence of filters (i.e. tuple)
            if isinstance(step, tuple):
                nbatch += 1
                desc = f"Batch {nbatch} / {nbatches} :" if verbose else None
                samples = self._fit(step, samples, executor, max_workers, desc)

            # Apply reducer on all samples in `state`
            else:
                samples = step.fit(samples)

        if verbose:
            end = time.time()
            print(f"\nProcessed samples in {end - start} s\n")

        return samples


    def steps(self):
        '''Return the order of processing steps to apply as a list where all
        consecutive sequences of filters are collapsed into tuples.

        E.g. [F, F, R, F, R, R, F, F, F] -> [(F, F), R, (F), R, R, (F, F, F)].
        '''
        start_filter = 0
        collecting_filters = False

        order = []

        for i, t in enumerate(self.transformers):
            if isinstance(t, Filter) and not collecting_filters:
                # Found the first filter in a new sequence of filters
                collecting_filters = True
                start_filter = i

            elif isinstance(t, Reducer):
                # Reducers are appended unchanged, unless we were collecting a
                # sequence of filters - in which case append filter sequence
                if collecting_filters:
                    filters_seq = tuple(self.transformers[start_filter:i])
                    order.append(filters_seq)

                    collecting_filters = False

                # Append reducer
                order.append(t)

        if collecting_filters:
            # Last running sequence of filters
            filters_seq = tuple(self.transformers[start_filter:])
            order.append(filters_seq)

        return order


    def __add__(self, other):
        if isinstance(other, Transformer):
            return Pipeline(self.transformers + [other])
        elif isinstance(other, Pipeline):
            return Pipeline(self.transformers + other.transformers)
        else:
            raise TypeError(textwrap.fill((
                "All new transformers added to the pipeline must be "
                "subclasses of `pept.base.Filter`, `pept.base.Reducer` or "
                f"`pept.base.Pipeline`. Received `{type(other)}`."
            )))


    def __repr__(self):
        name = type(self).__name__
        underline = "-" * len(name)
        return f"{name}\n{underline}\ntransformers = [\n" + textwrap.indent(
            "\n".join((t.__str__() for t in self.transformers)), '    '
        ) + "\n]"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.08.2019


'''PEPT-oriented visulisation tools.

Visualisation functions and classes for PEPT data, transparently working with
both `pept` base classes and raw NumPy arrays (e.g. `PlotlyGrapher.add_lines`
handles both `pept.LineData` and (N, 7) NumPy arrays).

The `PlotlyGrapher` class creates interactive, publication-ready 3D figures
with optional subplots which can also be exported to portable HTML files. The
`PlotlyGrapher2D` class is its two-dimensional counterpart, handling e.g.
`pept.Pixels`.

'''


from    .plotly_grapher     import  PlotlyGrapher
from    .plotly_grapher2d   import  PlotlyGrapher2D
from    .plotly_grapher2d   import  format_fig


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 31.01.2020


'''PEPT base classes.
'''


from    .line_data          import  LineData

from    .point_data         import  PointData

from    .pixel_data         import  Pixels

from    .voxel_data         import  Voxels
from    .voxel_data         import  VoxelData

from    .iterable_samples   import  IterableSamples
from    .iterable_samples   import  TimeWindow
from    .iterable_samples   import  AsyncIterableSamples
from    .iterable_samples   import  PEPTObject

from    .pipelines          import  Transformer
from    .pipelines          import  Filter
from    .pipelines          import  LineDataFilter
from    .pipelines          import  PointDataFilter
from    .pipelines          import  VoxelsFilter
from    .pipelines          import  Reducer
from    .pipelines          import  Pipeline




__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : line_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.08.2019


from    textwrap                import  indent

import  numpy                   as      np

import  matplotlib.pyplot       as      plt

from    .iterable_samples       import  IterableSamples




class LineData(IterableSamples):
    '''A class for PEPT LoR data iteration, manipulation and visualisation.

    Generally, PEPT Lines of Response (LoRs) are lines in 3D space, each
    defined by two points, regardless of the geometry of the scanner used. This
    class is used for the encapsulation of LoRs (or any lines!), efficiently
    yielding samples of `lines` of an adaptive `sample_size` and `overlap`.

    It is an abstraction over PET / PEPT scanner geometries and data formats,
    as once the raw LoRs (be they stored as binary, ASCII, etc.) are
    transformed into the common `LineData` format, any tracking, analysis or
    visualisation algorithm in the `pept` package can be used interchangeably.
    Moreover, it provides a stable, user-friendly interface for iterating over
    LoRs in *samples* - this is useful for tracking algorithms, as they
    generally take a few LoRs (a *sample*), produce a tracer position, then
    move to the next sample of LoRs, repeating the procedure. Using overlapping
    samples is also useful for improving the tracking rate of the algorithms.

    This is the base class for LoR data; the subroutines for transforming other
    data formats into `LineData` can be found in `pept.scanners`. If you'd like
    to integrate another scanner geometry or raw data format into this package,
    you can check out the `pept.scanners.parallel_screens` module as an
    example. This usually only involves writing a single function by hand; then
    all attributes and methods from `LineData` will be available to your new
    data format. If you'd like to use `LineData` as the base for other
    algorithms, you can check out the `pept.tracking.peptml.cutpoints` module
    as an example; the `Cutpoints` class iterates the samples of LoRs in any
    `LineData` **in parallel**, using `concurrent.futures.ThreadPoolExecutor`.

    Attributes
    ----------
    lines : (N, M>=7) numpy.ndarray
        An (N, M>=7) numpy array that stores the PEPT LoRs as time and
        cartesian (3D) coordinates of two points defining a line, followed by
        any additional data. The data columns are then
        `[time, x1, y1, z1, x2, y2, z2, etc.]`.

    sample_size : int, list[int], pept.TimeWindow or None
        Defining the number of LoRs in a sample; if it is an integer, a
        constant number of LoRs are returned per sample. If it is a list of
        integers, sample `i` will have length `sample_size[i]`. If it is a
        `pept.TimeWindow` instance, each sample will span a fixed time window.
        If `None`, custom sample sizes are returned as per the
        `samples_indices` attribute.

    overlap : int, pept.TimeWindow or None
        Defining the overlapping LoRs between consecutive samples. If `int`,
        constant numbers of LoRs are used. If `pept.TimeWindow`, the overlap
        will be a constant time window across the data timestamps (first
        column). If `None`, custom sample sizes are defined as per the
        `samples_indices` attribute.

    samples_indices : (S, 2) numpy.ndarray
        A 2D NumPy array of integers, where row `i` defines the i-th sample's
        start and end row indices, i.e.
        `sample[i] == data[samples_indices[i, 0]:samples_indices[i, 1]]`. The
        `sample_size` and `overlap` are simply friendly interfaces to setting
        the `samples_indices`.

    columns : (M,) list[str]
        A list of strings with the same number of columns as `lines` containing
        each column's name.

    attrs : dict[str, Any]
        A dictionary of other attributes saved on this class. Attribute names
        starting with an underscore are considered "hidden".

    Examples
    --------
    Initialise a `LineData` instance containing 10 lines with a `sample_size`
    of 3.

    >>> import pept
    >>> import numpy as np
    >>> lines_raw = np.arange(70).reshape(10, 7)
    >>> print(lines_raw)
    [[ 0  1  2  3  4  5  6]
     [ 7  8  9 10 11 12 13]
     [14 15 16 17 18 19 20]
     [21 22 23 24 25 26 27]
     [28 29 30 31 32 33 34]
     [35 36 37 38 39 40 41]
     [42 43 44 45 46 47 48]
     [49 50 51 52 53 54 55]
     [56 57 58 59 60 61 62]
     [63 64 65 66 67 68 69]]

    >>> line_data = pept.LineData(lines_raw, sample_size = 3)
    >>> line_data
    pept.LineData (samples: 3)
    --------------------------
    sample_size = 3
    overlap = 0
    lines =
      (rows: 10, columns: 7)
      [[ 0.  1. ...  5.  6.]
       [ 7.  8. ... 12. 13.]
       ...
       [56. 57. ... 61. 62.]
       [63. 64. ... 68. 69.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    Access samples using subscript notation. Notice how the samples are
    consecutive, as `overlap` is 0 by default.

    >>> line_data[0]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 3
    overlap = 0
    lines =
      (rows: 3, columns: 7)
      [[ 0.  1. ...  5.  6.]
       [ 7.  8. ... 12. 13.]
       [14. 15. ... 19. 20.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    >>> line_data[1]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 3
    overlap = 0
    lines =
      (rows: 3, columns: 7)
      [[21. 22. ... 26. 27.]
       [28. 29. ... 33. 34.]
       [35. 36. ... 40. 41.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    Now set an overlap of 2; notice how the number of samples changes:

    >>> len(line_data)     # Number of samples
    3

    >>> line_data.overlap = 2
    >>> len(line_data)
    8

    Notice how rows are repeated from one sample to the next when accessing
    them, because `overlap` is now 2:

    >>> line_data[0]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 3
    overlap = 0
    lines =
      (rows: 3, columns: 7)
      [[ 0.  1. ...  5.  6.]
       [ 7.  8. ... 12. 13.]
       [14. 15. ... 19. 20.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    >>> line_data[1]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 3
    overlap = 0
    lines =
      (rows: 3, columns: 7)
      [[ 7.  8. ... 12. 13.]
       [14. 15. ... 19. 20.]
       [21. 22. ... 26. 27.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    Now change `sample_size` to 5 and notice again how the number of samples
    changes:

    >>> len(line_data)
    8

    >>> line_data.sample_size = 5
    >>> len(line_data)
    2

    >>> line_data[0]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 5
    overlap = 0
    lines =
      (rows: 5, columns: 7)
      [[ 0.  1. ...  5.  6.]
       [ 7.  8. ... 12. 13.]
       ...
       [21. 22. ... 26. 27.]
       [28. 29. ... 33. 34.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    >>> line_data[1]
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 5
    overlap = 0
    lines =
      (rows: 5, columns: 7)
      [[21. 22. ... 26. 27.]
       [28. 29. ... 33. 34.]
       ...
       [42. 43. ... 47. 48.]
       [49. 50. ... 54. 55.]]
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']
    attrs = {}

    Notice how the samples do not cover the whole input `lines_raw` array, as
    the last lines are omitted - think of the `sample_size` and `overlap`. They
    are still inside the inner `lines` attribute of `line_data` though:

    >>> line_data.lines
    array([[ 0.,  1.,  2.,  3.,  4.,  5.,  6.],
           [ 7.,  8.,  9., 10., 11., 12., 13.],
           [14., 15., 16., 17., 18., 19., 20.],
           [21., 22., 23., 24., 25., 26., 27.],
           [28., 29., 30., 31., 32., 33., 34.],
           [35., 36., 37., 38., 39., 40., 41.],
           [42., 43., 44., 45., 46., 47., 48.],
           [49., 50., 51., 52., 53., 54., 55.],
           [56., 57., 58., 59., 60., 61., 62.],
           [63., 64., 65., 66., 67., 68., 69.]])

    Notes
    -----
    The class saves `lines` as a **C-contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the number of rows or columns after instantiating the class.

    See Also
    --------
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    pept.tracking.Cutpoints : Compute cutpoints from `pept.LineData`.
    '''

    def __init__(
        self,
        lines,
        sample_size = None,
        overlap = None,
        columns = ["t", "x1", "y1", "z1", "x2", "y2", "z2"],
        **kwargs,
    ):
        '''`LineData` class constructor.

        Parameters
        ----------
        lines : (N, M>=7) numpy.ndarray
            An (N, M>=7) numpy array that stores the PEPT LoRs (or any generic
            set of lines) as time and cartesian (3D) coordinates of two points
            defining each line, followed by any additional data. The data
            columns are then `[time, x1, y1, z1, x2, y2, z2, etc.]`.

        sample_size : int, default 0
            An `int` that defines the number of lines that should be returned
            when iterating over `lines`. A `sample_size` of 0 yields all the
            data as one single sample.

        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `lines`. An overlap of 0
            means consecutive samples, while an overlap of (`sample_size` - 1)
            means incrementing the samples by one. A negative overlap means
            skipping values between samples. An error is raised if `overlap` is
            larger than or equal to `sample_size`.

        columns : List[str], default ["t", "x1", "y1", "z1", "x2", "y2", "z2"]
            A list of strings corresponding to the column labels in `points`.

        **kwargs : extra keyword arguments
            Any extra attributes to set in `.attrs`.

        Raises
        ------
        ValueError
            If `lines` has fewer than 7 columns.

        ValueError
            If `overlap` >= `sample_size` unless `sample_size` is 0. Overlap
            has to be smaller than `sample_size`. Note that it can also be
            negative.
        '''

        # Copy-constructor
        if isinstance(lines, LineData):
            kwargs.update(lines.attrs)
            columns = lines.columns
            lines = lines.lines.copy()

        # Iterable of LineData
        if len(lines) and isinstance(lines[0], LineData):
            kwargs.update(lines[0].attrs)
            columns = lines[0].columns
            sample_size = [len(li.lines) for li in lines]
            lines = np.vstack([li.lines for li in lines])

        # NumPy array-like
        else:
            lines = np.asarray(lines, order = 'C', dtype = float)

        # Check that lines has at least 7 columns.
        if lines.ndim != 2 or lines.shape[1] < 7:
            raise ValueError((
                "\n[ERROR]: `lines` should have dimensions (M, N), where "
                f"N >= 7. Received {lines.shape}.\n"
            ))

        # Call the IterableSamples constructor to make the class iterable in
        # samples with overlap.
        IterableSamples.__init__(
            self,
            lines,
            sample_size,
            overlap,
            columns = columns,
            **kwargs,
        )


    @property
    def lines(self):
        # The `data` attribute is set by the parent class, `IterableSamples`
        return self.data


    def to_csv(self, filepath, delimiter = " "):
        '''Write `lines` to a CSV file.

        Write all LoRs stored in the class to a CSV file.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        delimiter : str, default " "
            The delimiter used to separate the values in the CSV file.

        '''

        np.savetxt(filepath, self.lines, delimiter = delimiter,
                   header = delimiter.join(self.columns))


    def plot(
        self,
        sample_indices = ...,
        ax = None,
        alt_axes = False,
        colorbar_col = 0,
    ):
        '''Plot lines from selected samples using matplotlib.

        Returns matplotlib figure and axes objects containing all lines
        included in the samples selected by `sample_indices`.
        `sample_indices` may be a single sample index (e.g. 0), an iterable
        of indices (e.g. [1,5,6]), or an Ellipsis (`...`) for all samples.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of lines. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all lines).

        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        alt_axes : bool, default False
            If `True`, plot using the alternative PEPT-style axes convention:
            z is horizontal, y points upwards. Because Matplotlib cannot swap
            axes, this is achieved by swapping the parameters in the plotting
            call (i.e. `plt.plot(x, y, z)` -> `plt.plot(z, x, y)`).

        colorbar_col : int, default -1
            The column in the data samples that will be used to color the
            lines. The default is -1 (the last column).

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Notes
        -----
        Plotting all lines is very computationally-expensive for matplotlib. It
        is recommended to only plot a couple of samples at a time, or use the
        faster `pept.plots.PlotlyGrapher`.

        Examples
        --------
        Plot the lines from sample 1 in a `LineData` instance:

        >>> lors = pept.LineData(...)
        >>> fig, ax = lors.plot(1)
        >>> fig.show()

        Plot the lines from samples 0, 1 and 2:

        >>> fig, ax = lors.plot([0, 1, 2])
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection = '3d')
        else:
            fig = plt.gcf()

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        lines, color = [], []
        # For each selected sample include all the lines' coordinates
        for n in sample_indices:
            # If an Ellipsis was received, then include all lines.
            if n is Ellipsis:
                sample = self.lines
            else:
                sample = self[n]

            for line in sample:
                lines.append(line[0:7])
                color.append(line[colorbar_col])

        color = np.array(color)

        # Scatter x, y, z [color]
        cmap = plt.cm.magma
        color_array = cmap(color / color.max())

        if alt_axes:
            for i, line in enumerate(lines):
                ax.plot(
                    [line[3], line[6]],
                    [line[1], line[4]],
                    [line[2], line[5]],
                    c = color_array[i],
                    alpha = 0.8
                )

            ax.set_xlabel("z (mm)")
            ax.set_ylabel("x (mm)")
            ax.set_zlabel("y (mm)")

        else:
            for i, line in enumerate(lines):
                ax.plot(
                    [line[1], line[4]],
                    [line[2], line[5]],
                    [line[3], line[6]],
                    c = color_array[i],
                    alpha = 0.8
                )

            ax.set_xlabel("x (mm)")
            ax.set_ylabel("y (mm)")
            ax.set_zlabel("z (mm)")

        return fig, ax


    def __repr__(self):
        # String representation of the class
        name = f"pept.LineData (samples: {len(self)})"
        underline = "-" * len(name)

        # Custom printing of the .lines and .samples_indices arrays
        with np.printoptions(threshold = 5, edgeitems = 2):
            lines_str = f"{indent(str(self.lines), '  ')}"

            if self.sample_size is None:
                samples_indices_str = str(self.samples_indices)
                samples_indices_str = (
                    f"samples_indices = \n"
                    f"{indent(samples_indices_str, '  ')}\n"
                )
            else:
                samples_indices_str = ""

        # Pretty-printing extra attributes
        attrs_str = ""
        if self.attrs:
            items = []
            for k, v in self.attrs.items():
                s = f"  {k.__repr__()}: {v}"
                if len(s) > 75:
                    s = s[:72] + "..."
                items.append(s)
            attrs_str = "\n" + "\n".join(items) + "\n"

        # Return constructed string
        return (
            f"{name}\n{underline}\n"
            f"sample_size = {self.sample_size}\n"
            f"overlap = {self.overlap}\n"
            f"{samples_indices_str}"
            f"lines = \n"
            f"  (rows: {len(self.lines)}, columns: {len(self.columns)})\n"
            f"{lines_str}\n"
            f"columns = {self.columns}\n"
            "attrs = {"
            f"{attrs_str}"
            "}\n"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : space_transformers.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 11.08.2021


import  textwrap

import  numpy               as      np
from    scipy.interpolate   import  interp1d

from    pept                import  LineData, PointData, Voxels
from    pept.base           import  LineDataFilter, PointDataFilter




class Voxelize(LineDataFilter):
    '''Asynchronously voxelize samples of lines from a `pept.LineData`.

    Filter signature:

    ::

        LineData -> Voxelize.fit_sample -> PointData

    This filter is much more memory-efficient than voxelizing all samples of
    LoRs at once - which often overflows the available memory. Most often this
    is used alongside voxel-based tracking algorithms, e.g.
    ``pept.tracking.FPI``:

    >>> from pept.tracking import *
    >>> pipeline = pept.Pipeline([
    >>>     Voxelize((50, 50, 50)),
    >>>     FPI(3, 0.4),
    >>>     Stack(),
    >>> ])

    Parameters
    ----------
    number_of_voxels : 3-tuple
        A tuple-like containing exactly three integers specifying the number of
        voxels to be used in each dimension.

    xlim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        x-dimension, formatted as [x_min, x_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    ylim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        y-dimension, formatted as [y_min, y_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    zlim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        z-dimension, formatted as [z_min, z_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    set_lines : (N, 7) numpy.ndarray or pept.LineData, optional
        If defined, set the system limits upon creating the class to the
        bounding box of the lines in `set_lines`.

    '''

    def __init__(
        self,
        number_of_voxels,
        xlim = None,
        ylim = None,
        zlim = None,
        set_lims = None,
    ):
        # Type-checking inputs
        number_of_voxels = np.asarray(
            number_of_voxels,
            order = "C",
            dtype = int
        )

        if number_of_voxels.ndim != 1 or len(number_of_voxels) != 3:
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of voxels in the x-, y-, and z-dimension. "
                f"Received parameter with shape {number_of_voxels.shape}."
            )))

        if (number_of_voxels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must set at least two "
                "voxels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_voxels}`."
            )))

        # Keep track of limits we have to set
        set_xlim = True
        set_ylim = True
        set_zlim = True

        if xlim is not None:
            set_xlim = False
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the x-dimension. "
                    f"Received parameter with shape {xlim.shape}."
                )))

        if ylim is not None:
            set_ylim = False
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the y-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        if zlim is not None:
            set_zlim = False
            zlim = np.asarray(zlim, dtype = float)

            if zlim.ndim != 1 or len(zlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `zlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the z-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        # Setting class attributes
        self._number_of_voxels = tuple(number_of_voxels)
        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim

        if set_lims is not None:
            if isinstance(set_lims, LineData):
                self.set_lims(set_lims.lines, set_xlim, set_ylim, set_zlim)
            else:
                self.set_lims(set_lims, set_xlim, set_ylim, set_zlim)


    def set_lims(
        self,
        lines,
        set_xlim = True,
        set_ylim = True,
        set_zlim = True,
    ):
        lines = np.asarray(lines, dtype = float, order = "C")

        if set_xlim:
            xlim = Voxels.get_cutoff(lines[:, 1], lines[:, 4])

        if set_ylim:
            ylim = Voxels.get_cutoff(lines[:, 2], lines[:, 5])

        if set_zlim:
            zlim = Voxels.get_cutoff(lines[:, 3], lines[:, 6])

        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim


    @property
    def number_of_voxels(self):
        return self._number_of_voxels


    @property
    def xlim(self):
        return self._xlim


    @property
    def ylim(self):
        return self._ylim


    @property
    def zlim(self):
        return self._zlim


    def fit_sample(self, sample_lines):
        if not isinstance(sample_lines, LineData):
            sample_lines = LineData(sample_lines)

        vox = Voxels.from_lines(
            sample_lines.lines,
            self.number_of_voxels,
            self.xlim,
            self.ylim,
            self.zlim,
            verbose = False,
        )

        # Save the constituent LoRs as a hidden attribute
        vox.attrs["_lines"] = sample_lines
        return vox




class Interpolate(PointDataFilter):
    '''Interpolate between data points at a fixed sampling rate; useful for
    Eulerian fields computation.

    Filter signature:

    ::

        PointData -> Interpolate.fit_sample -> PointData

    By default, the linear interpolator `scipy.interpolate.interp1d` is used.
    You can specify a different interpolator and keyword arguments to send it.
    E.g. nearest-neighbour interpolation: ``Interpolate(1., kind = "nearest")``
    or cubic interpolation: ``Interpolate(1., kind = "cubic")``.

    All data columns except timestamps are interpolated.
    '''

    def __init__(self, timestep, interpolator = interp1d, **kwargs):
        self.timestep = float(timestep)
        self.interpolator = interpolator
        self.kwargs = kwargs


    def fit_sample(self, sample):
        if not isinstance(sample, PointData):
            sample = PointData(sample)

        if not len(sample):
            return sample[0:0]

        # Create interpolators for each dimension (column) wrt time
        interps = [
            self.interpolator(
                sample.points[:, 0],
                sample.points[:, i],
                **self.kwargs,
            ) for i in range(1, sample.points.shape[1])
        ]

        # Sampling points, between the first and last timestamps
        sampling = np.arange(sample.points[0, 0], sample.points[-1, 0],
                             self.timestep)

        data = np.vstack([sampling] + [interp(sampling) for interp in interps])
        return sample.copy(data = data.T)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019 Andrei Leonard Nicusan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : modular_camera.py
# License: License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


import  time

import  numpy                       as      np

from    pept                        import  LineData
from    .extensions.get_pept_event  import  get_pept_LOR


def modular_camera(
    data_file,
    sample_size = None,
    overlap = None,
    verbose = True
):
    '''Initialise PEPT LoRs from the modular camera DAQ.

    Can read data from a `.da_1` file or equivalent. The file must contain
    the standard datawords from the modular camera output. This will then
    be automatically transformed into the standard `LineData` format
    with every row being `[time, x1, y1, z1, x2, y2, z2]`, where the geometry
    is derived from the C-extension. The current useable geometry is a square
    layout with 4 stacks for 4 modules, separated by 250 mm.

    Parameters
    ----------
    data_file : str
        A string with the (absolute or relative) path to the data file
        from which the PEPT data will be read. It should include the
        full file name, along with the extension (.da_1)

    sample_size : int, optional
        An `int`` that defines the number of lines that should be
        returned when iterating over `_lines`. A `sample_size` of 0
        yields all the data as one single sample. (Default is 200)

    overlap : int, optional
        An `int` that defines the overlap between two consecutive
        samples that are returned when iterating over `_lines`.
        An overlap of 0 means consecutive samples, while an overlap
        of (`sample_size` - 1) means incrementing the samples by one.
        A negative overlap means skipping values between samples. An
        error is raised if `overlap` is larger than or equal to
        `sample_size`. (Default is 0)

    verbose : bool, optional
        An option that enables printing the time taken for the
        initialisation of an instance of the class. Useful when
        reading large files (10gb files for PEPT data is not unheard
        of). (Default is True)

    Returns
    -------
    LineData
        The initialised LoRs.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size`. Overlap has to be smaller than
        `sample_size`. Note that it can also be negative.

    ValueError
        If the data file does not have (N, 7) shape.
    '''

    if verbose:
        start = time.time()

    x = 10

    header_buffer_size = 1000

    n_events = 0

    # Modular camera data reader requires 'itag' for timing. We will drop
    # this column at the end of initialisation
    # Row: [itag, itime, X1, Y1, Z1, X2, Y2, Z2]

    lines = np.zeros([sample_size, 8])

    with open(data_file, "rb") as f:
        # Skip over the header and handshake word
        f.seek(header_buffer_size)

        word = f.read(4)

        if word.hex() == 'cefacefa':
            # Skip two words
            word = f.read(4)
            word = f.read(4)

        itime = 0
        itag = 0

        # BufTime = 0
        # nBuf = 0

        while word != b'' and (n_events < sample_size):

            word = f.read(4)

            # Handshake word
            if word.hex() == 'cefacefa':
                # Skip two words
                word = f.read(4)
                word = f.read(4)

            if word != b'':
                word = int.from_bytes(word, "little")

                lines[n_events, :] = get_pept_LOR(
                    word, itag, itime
                )   # C function

                itag = lines[n_events, 1]
                itime = lines[n_events, 1]

                n_events = n_events + 1

                if (n_events % x) == 0:
                    print("Got ", n_events, "\n")
                    x = x * 10

    # Remove 'zero' lines
    lines = lines[np.all(lines, axis = 1)]

    # Drop itag column
    lines = np.delete(lines, 0, axis=1)

    if verbose:
        end = time.time()
        print(f"\nInitialised the PEPT data in {end - start:3.3f} s.\n")

    return LineData(lines, sample_size = sample_size, overlap = overlap)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : point_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.08.2019


from    textwrap                import  indent

import  numpy                   as      np

import  matplotlib.pyplot       as      plt

from    .iterable_samples       import  IterableSamples




class PointData(IterableSamples):
    '''A class for general PEPT point-like data iteration, manipulation and
    visualisation.

    In the context of positron-based particle tracking, points are defined by a
    timestamp, 3D coordinates and any other extra information (such as
    trajectory label or some tracer signature). This class is used for the
    encapsulation of 3D points - be they tracer locations, cutpoints, etc. -,
    efficiently yielding samples of `points` of an adaptive `sample_size` and
    `overlap`.

    Much like a complement to `LineData`, `PointData` is an abstraction over
    point-like data that may be encountered in the context of PEPT (e.g.
    pre-tracked tracer locations), as once the raw points are transformed into
    the common `PointData` format, any tracking, analysis or visualisation
    algorithm in the `pept` package can be used interchangeably. Moreover, it
    provides a stable, user-friendly interface for iterating over points in
    *samples* - this can be useful for tracking algorithms, as some take a few
    points (a *sample*), produce an accurate tracer location, then move to the
    next sample of points, repeating the procedure. Using overlapping samples
    is also useful for improving the time resolution of the algorithms.

    This is the base class for point-like data; subroutines that accept and/or
    return `PointData` instances (or subclasses thereof) can be found
    throughout the `pept` package. If you'd like to create new algorithms based
    on them, you can check out the `pept.tracking.peptml.cutpoints` module as
    an example; the `Cutpoints` class receives a `LineData` instance,
    transforms the samples of LoRs into cutpoints, then initialises itself as a
    `PointData` subclass - thereby inheriting all its methods and attributes.

    Attributes
    ----------
    points : (N, M) numpy.ndarray
        An (N, M >= 4) numpy array that stores the points as time, followed by
        cartesian (3D) coordinates of the point, followed by any extra
        information. The data columns are then `[time, x, y, z, etc]`.

    sample_size : int, list[int], pept.TimeWindow or None
        Defining the number of points in a sample; if it is an integer, a
        constant number of points are returned per sample. If it is a list of
        integers, sample `i` will have length `sample_size[i]`. If it is a
        `pept.TimeWindow` instance, each sample will span a fixed time window.
        If `None`, custom sample sizes are returned as per the
        `samples_indices` attribute.

    overlap : int, pept.TimeWindow or None
        Defining the overlapping points between consecutive samples. If `int`,
        constant numbers of points are used. If `pept.TimeWindow`, the overlap
        will be a constant time window across the data timestamps (first
        column). If `None`, custom sample sizes are defined as per the
        `samples_indices` attribute.

    samples_indices : (S, 2) numpy.ndarray
        A 2D NumPy array of integers, where row `i` defines the i-th sample's
        start and end row indices, i.e.
        `sample[i] == data[samples_indices[i, 0]:samples_indices[i, 1]]`. The
        `sample_size` and `overlap` are simply friendly interfaces to setting
        the `samples_indices`.

    columns : (M,) list[str]
        A list of strings with the same number of columns as `points`
        containing each column's name.

    attrs : dict[str, Any]
        A dictionary of other attributes saved on this class. Attribute names
        starting with an underscore are considered "hidden".

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size`. Overlap is required to be smaller than
        `sample_size`, unless `sample_size` is 0. Note that it can also be
        negative.

    Notes
    -----
    This class saves `points` as a **C-contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the number of rows or columns after instantiating the class.

    Examples
    --------
    Initialise a `PointData` instance containing 10 points with a `sample_size`
    of 3.

    >>> import numpy as np
    >>> import pept
    >>> points_raw = np.arange(40).reshape(10, 4)
    >>> print(points_raw)
    [[ 0  1  2  3]
     [ 4  5  6  7]
     [ 8  9 10 11]
     [12 13 14 15]
     [16 17 18 19]
     [20 21 22 23]
     [24 25 26 27]
     [28 29 30 31]
     [32 33 34 35]
     [36 37 38 39]]

    >>> point_data = pept.PointData(points_raw, sample_size = 3)
    >>> point_data
    pept.PointData (samples: 3)
    ---------------------------
    sample_size = 3
    overlap = 0
    points =
      (rows: 10, columns: 4)
      [[ 0.  1.  2.  3.]
       [ 4.  5.  6.  7.]
       ...
       [32. 33. 34. 35.]
       [36. 37. 38. 39.]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {}

    Access samples using subscript notation. Notice how the samples are
    consecutive, as `overlap` is 0 by default.

    >>> point_data[0]
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 3
    overlap = 0
    points =
      (rows: 3, columns: 4)
      [[ 0.  1.  2.  3.]
       [ 4.  5.  6.  7.]
       [ 8.  9. 10. 11.]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {}

    >>> point_data[1]
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 3
    overlap = 0
    points =
      (rows: 3, columns: 4)
      [[12. 13. 14. 15.]
       [16. 17. 18. 19.]
       [20. 21. 22. 23.]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {}

    Now set an overlap of 2; notice how the number of samples changes:

    >>> len(point_data)         # Number of samples
    3

    >>> point_data.overlap = 2
    >>> len(point_data)
    8

    Notice how rows are repeated from one sample to the next when accessing
    them, because `overlap` is now 2:

    >>> point_data[0]
    array([[ 0.,  1.,  2.,  3.],
           [ 4.,  5.,  6.,  7.],
           [ 8.,  9., 10., 11.]])

    >>> point_data[1]
    array([[ 4.,  5.,  6.,  7.],
           [ 8.,  9., 10., 11.],
           [12., 13., 14., 15.]])

    Now change `sample_size` to 5 and notice again how the number of samples
    changes:

    >>> len(point_data)
    8

    >>> point_data.sample_size = 5
    >>> len(point_data)
    2

    >>> point_data[0]
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 3
    overlap = 0
    points =
      (rows: 3, columns: 4)
      [[ 0.  1.  2.  3.]
       [ 4.  5.  6.  7.]
       [ 8.  9. 10. 11.]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {}

    >>> point_data[1]
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 3
    overlap = 0
    points =
      (rows: 3, columns: 4)
      [[ 4.  5.  6.  7.]
       [ 8.  9. 10. 11.]
       [12. 13. 14. 15.]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {}

    Notice how the samples do not cover the whole input `points_raw` array, as
    the last lines are omitted - think of the `sample_size` and `overlap`. They
    are still inside the inner `points` attribute of `point_data` though:

    >>> point_data.points
    array([[ 0.,  1.,  2.,  3.],
           [ 4.,  5.,  6.,  7.],
           [ 8.,  9., 10., 11.],
           [12., 13., 14., 15.],
           [16., 17., 18., 19.],
           [20., 21., 22., 23.],
           [24., 25., 26., 27.],
           [28., 29., 30., 31.],
           [32., 33., 34., 35.],
           [36., 37., 38., 39.]])

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.read_csv : Fast CSV file reading into numpy arrays.
    pept.plots.PlotlyGrapher :
        Easy, publication-ready plotting of PEPT-oriented data.
    pept.tracking.Cutpoints : Compute cutpoints from `pept.LineData`.
    '''

    def __init__(
        self,
        points,
        sample_size = None,
        overlap = None,
        columns = ["t", "x", "y", "z"],
        **kwargs,
    ):
        '''`PointData` class constructor.

        Parameters
        ----------
        points : (N, M) numpy.ndarray
            An (N, M >= 4) numpy array that stores points (or any generic 2D
            set of data). It expects that the first column is time, followed by
            cartesian (3D) coordinates of points, followed by any extra
            information the user needs. The data columns are then
            `[time, x, y, z, etc]`.

        sample_size : int, default 0
            An `int`` that defines the number of points that should be returned
            when iterating over `points`. A `sample_size` of 0 yields all the
            data as one single sample.

        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `points`. An overlap of 0
            means consecutive samples, while an overlap of (`sample_size` - 1)
            implies incrementing the samples by one. A negative overlap means
            skipping values between samples. An error is raised if `overlap` is
            larger than or equal to `sample_size`.

        columns : List[str], default ["t", "x", "y", "z"]
            A list of strings corresponding to the column labels in `points`.

        **kwargs : extra keyword arguments
            Any extra attributes to set on the class instance.

        Raises
        ------
        ValueError
            If `line_data` does not have (N, M) shape, where M >= 4.
        '''

        # Copy-constructor
        if isinstance(points, PointData):
            kwargs.update(points.attrs)
            columns = points.columns
            points = points.points.copy()

        # Iterable of PointData
        if len(points) and isinstance(points[0], PointData):
            kwargs.update(points[0].attrs)
            columns = points[0].columns
            sample_size = [len(p.points) for p in points]
            points = np.vstack([p.points for p in points])

        # NumPy array-like
        else:
            points = np.asarray(points, order = 'C', dtype = float)

        # Check that points has at least 4 columns.
        if points.ndim != 2 or points.shape[1] < 4:
            raise ValueError((
                "\n[ERROR]: `points` should have dimensions (M, N), where "
                f"N >= 4. Received {points.shape}.\n"
            ))

        # Call the IterableSamples constructor to make the class iterable in
        # samples with overlap.
        IterableSamples.__init__(
            self,
            points,
            sample_size,
            overlap,
            columns = columns,
            **kwargs,
        )


    @property
    def points(self):
        # The `data` attribute is set by the parent class, `IterableSamples`
        return self.data


    def to_csv(self, filepath, delimiter = " "):
        '''Write the inner `points` to a CSV file.

        Write all points stored in the class to a CSV file.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        delimiter : str, default " "
            The delimiter used to separate the values in the CSV file.

        '''

        np.savetxt(filepath, self.points, delimiter = delimiter,
                   header = delimiter.join(self.columns))


    def plot(
        self,
        sample_indices = ...,
        ax = None,
        alt_axes = False,
        colorbar_col = -1,
    ):
        '''Plot points from selected samples using matplotlib.

        Returns matplotlib figure and axes objects containing all points
        included in the samples selected by `sample_indices`.
        `sample_indices` may be a single sample index (e.g. 0), an iterable
        of indices (e.g. [1,5,6]), or an Ellipsis (`...`) for all samples.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of points. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all points).

        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        alt_axes : bool, default False
            If `True`, plot using the alternative PEPT-style axes convention:
            z is horizontal, y points upwards. Because Matplotlib cannot swap
            axes, this is achieved by swapping the parameters in the plotting
            call (i.e. `plt.plot(x, y, z)` -> `plt.plot(z, x, y)`).

        colorbar_col : int, default -1
            The column in the data samples that will be used to color the
            points. The default is -1 (the last column).

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Notes
        -----
        Plotting all points is very computationally-expensive for matplotlib.
        It is recommended to only plot a couple of samples at a time, or use
        the faster `pept.plots.PlotlyGrapher`.

        Examples
        --------
        Plot the points from sample 1 in a `PointData` instance:

        >>> point_data = pept.PointData(...)
        >>> fig, ax = point_data.plot(1)
        >>> fig.show()

        Plot the points from samples 0, 1 and 2:

        >>> fig, ax = point_data.plot([0, 1, 2])
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection = '3d')
        else:
            fig = plt.gcf()

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        if sample_indices[0] == Ellipsis:
            x = self.points[:, 1],
            y = self.points[:, 2],
            z = self.points[:, 3],
            color = self.points[:, colorbar_col]
        else:
            x, y, z, color = [], [], [], []
            for n in sample_indices:
                sample = self[n]

                x.extend(sample[:, 1])
                y.extend(sample[:, 2])
                z.extend(sample[:, 3])

                color.extend(sample[:, colorbar_col])

        color = np.array(color)

        # Scatter x, y, z, [color]
        cmap = plt.cm.magma
        color_array = cmap(color / color.max())

        if alt_axes:
            ax.scatter(z, x, y, c = color_array)

            ax.set_xlabel("z (mm)")
            ax.set_ylabel("x (mm)")
            ax.set_zlabel("y (mm)")

        else:
            ax.scatter(x, y, z, c = color_array)

            ax.set_xlabel("x (mm)")
            ax.set_ylabel("y (mm)")
            ax.set_zlabel("z (mm)")

        return fig, ax


    def __repr__(self):
        # String representation of the class
        name = f"pept.PointData (samples: {len(self)})"
        underline = "-" * len(name)

        # Custom printing of the .points and .samples_indices arrays
        with np.printoptions(threshold = 5, edgeitems = 2):
            points_str = f"{indent(str(self.points), '  ')}"

            if self.sample_size is None:
                samples_indices_str = str(self.samples_indices)
                samples_indices_str = (
                    f"samples_indices = \n"
                    f"{indent(samples_indices_str, '  ')}\n"
                )
            else:
                samples_indices_str = ""

        # Pretty-printing extra attributes
        attrs_str = ""
        if self.attrs:
            items = []
            for k, v in self.attrs.items():
                s = f"  {k.__repr__()}: {v}"
                if len(s) > 75:
                    s = s[:72] + "..."
                items.append(s)
            attrs_str = "\n" + "\n".join(items) + "\n"

        # Return constructed string
        return (
            f"{name}\n{underline}\n"
            f"sample_size = {self.sample_size}\n"
            f"overlap = {self.overlap}\n"
            f"{samples_indices_str}"
            f"points = \n"
            f"  (rows: {len(self.points)}, columns: {len(self.columns)})\n"
            f"{points_str}\n"
            f"columns = {self.columns}\n"
            "attrs = {"
            f"{attrs_str}"
            "}\n"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : birmingham_method.py
# License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019



from    beartype                        import  beartype
import  numpy                           as      np

import  pept

from    .extensions.birmingham_method   import  birmingham_method


class BirminghamMethod(pept.base.LineDataFilter):
    '''The Birmingham Method is an efficient, analytical technique for tracking
    tracers using the LoRs from PEPT data.

    Two main methods are provided: `fit_sample` for tracking a single numpy
    array of LoRs (i.e. a single sample) and `fit` which tracks all the samples
    encapsulated in a `pept.LineData` class *in parallel*.

    For the given `sample` of LoRs (a numpy.ndarray), this function minimises
    the distance between all of the LoRs, rejecting a fraction of lines that
    lie furthest away from the calculated distance. The process is repeated
    iteratively until a specified fraction (`fopt`) of the original subset of
    LORs remains.

    This class is a wrapper around the `birmingham_method` subroutine
    (implemented in C), providing tools for asynchronously tracking samples of
    LoRs. It can return `PointData` classes which can be easily manipulated and
    visualised.

    Attributes
    ----------
    fopt : float
        Floating-point number between 0 and 1, representing the target fraction
        of LoRs in a sample used to locate a tracer.

    get_used : bool, default False
        If True, attach an attribute ``._lines`` to the output PointData
        containing the sample of LoRs used (+ a column `used`).

    Examples
    --------
    A typical workflow would involve reading LoRs from a file, instantiating a
    `BirminghamMethod` class, tracking the tracer locations from the LoRs, and
    plotting them.

    >>> import pept
    >>> from pept.tracking.birmingham_method import BirminghamMethod

    >>> lors = pept.LineData(...)   # set sample_size and overlap appropriately
    >>> bham = BirminghamMethod()
    >>> locations = bham.fit(lors)  # this is a `pept.PointData` instance

    >>> grapher = PlotlyGrapher()
    >>> grapher.add_points(locations)
    >>> grapher.show()

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    pept.scanners.ParallelScreens : Initialise a `pept.LineData` instance from
                                    parallel screens PEPT detectors.
    '''

    def __init__(self, fopt = 0.5, get_used = False):
        '''`BirminghamMethod` class constructor.

        fopt : float, default 0.5
            Float number between 0 and 1, representing the fraction of
            remaining LORs in a sample used to locate the particle.

        verbose : bool, default False
            Print extra information when initialising this class.
        '''

        # Use @fopt.setter (below) to do the relevant type-checking when
        # setting fopt (self._fopt is the internal attribute, that we only
        # access through the getter and setter of the self.fopt property).
        self.fopt = float(fopt)
        self.get_used = bool(get_used)


    @beartype
    def fit_sample(self, sample):
        '''Use the Birmingham method to track a tracer location from a numpy
        array (i.e. one sample) of LoRs.

        For the given `sample` of LoRs (a numpy.ndarray), this function
        minimises the distance between all of the LoRs, rejecting a fraction of
        lines that lie furthest away from the calculated distance. The process
        is repeated iteratively until a specified fraction (`fopt`) of the
        original subset of LORs remains.

        Parameters
        ----------
        sample : (N, M>=7) numpy.ndarray
            The sample of LORs that will be clustered. Each LoR is expressed as
            a timestamps and a line defined by two points; the data columns are
            then `[time, x1, y1, z1, x2, y2, z2, extra...]`.

        get_used : bool, default False
            If `True`, the function will also return a boolean mask of the LoRs
            used to compute the tracer location - that is, a vector of the same
            length as `sample`, containing 1 for the rows that were used, and 0
            otherwise.

        as_array : bool, default True
            If set to True, the tracked locations are returned as numpy arrays.
            If set to False, they are returned inside an instance of
            `pept.PointData` for ease of iteration and plotting.

        verbose : bool, default False
            Provide extra information when tracking a location: time the
            operation and show a progress bar.

        Returns
        -------
        locations : numpy.ndarray or pept.PointData
            The tracked locations found.

        used : numpy.ndarray, optional
            If `get_used` is true, then also return a boolean mask of the LoRs
            used to compute the tracer location - that is, a vector of the same
            length as `sample`, containing 1 for the rows that were used, and 0
            otherwise.
            [ Used for multi-particle tracking, not implemented yet]

        Raises
        ------
        ValueError
            If `sample` is not a numpy array of shape (N, M), where M >= 7.
        '''

        if not isinstance(sample, pept.LineData):
            sample = pept.LineData(sample)

        locations, used = birmingham_method(sample.lines, self.fopt)

        # Propagate any LineData attributes besides `columns`
        attrs = sample.extra_attrs()

        locations = pept.PointData(
            [locations],
            columns = ["t", "x", "y", "z", "error"],
            **attrs,
        )

        # If `get_used`, also attach a `._lines` attribute with the lines used
        if self.get_used:
            locations.attrs["_lines"] = sample.copy(
                data = np.c_[sample.lines, used],
                columns = sample.columns + ["used"],
            )

        return locations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : post.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 06.09.2021


import  warnings

import  numpy           as      np

try:
    class NotInstalled:
        '''Class used to signal that Numba is not available.'''
        pass

    import  numba       as      nb
except ImportError:
    nb = NotInstalled()
    nb.njit = lambda func: func

from    pept            import  PointData
from    pept.base       import  PointDataFilter


@nb.njit
def polyfit(x, y, deg):
    '''Fit polynomial of order `deg` against x, y data points.'''
    mat = np.zeros(shape = (x.shape[0], deg + 1))
    mat[:, 0] = np.ones_like(x)
    for n in range(1, deg + 1):
        mat[:, n] = x**n

    p = np.linalg.lstsq(mat, y)[0]
    return p


@nb.njit
def polyder(p):
    '''Differentiate polynomial p.'''
    d = np.zeros(shape = (p.shape[0] - 1))
    for n in range(d.shape[0]):
        d[n] = (n + 1) * p[n + 1]
    return d


@nb.njit
def polyval(p, x):
    '''Evaluate polynomial p(x) using Horner's Method.

    New numpy.polynomial.Polynomial format:
        p[0] + p[1] * x + p[2] * x^2 + ...
    '''
    result = np.zeros_like(x)
    for coeff in p[::-1]:
        result = x * result + coeff
    return result


@nb.njit
def compute_velocities(points, window, deg):
    # Pre-allocate velocities matrix, columns [vx, vy, vz]
    v = np.zeros((points.shape[0], 3))

    # Half-window size
    hw = (window - 1) // 2

    # Infer velocities of first hw points
    w = points[:window]
    vx = polyder(polyfit(w[:, 0], w[:, 1], deg))
    vy = polyder(polyfit(w[:, 0], w[:, 2], deg))
    vz = polyder(polyfit(w[:, 0], w[:, 3], deg))

    for i in range(hw + 1):
        v[i, 0] = polyval(vx, w[i, 0:1])[0]
        v[i, 1] = polyval(vy, w[i, 0:1])[0]
        v[i, 2] = polyval(vz, w[i, 0:1])[0]

    # Compute velocities in a sliding window
    for i in range(hw + 1, points.shape[0] - hw):
        w = points[i - hw:i + hw + 1]

        vx = polyder(polyfit(w[:, 0], w[:, 1], deg))
        vy = polyder(polyfit(w[:, 0], w[:, 2], deg))
        vz = polyder(polyfit(w[:, 0], w[:, 3], deg))

        v[i, 0] = polyval(vx, points[i, 0:1])[0]
        v[i, 1] = polyval(vy, points[i, 0:1])[0]
        v[i, 2] = polyval(vz, points[i, 0:1])[0]

    # Infer velocities of last hw points
    for i in range(points.shape[0] - hw, points.shape[0]):
        v[i, 0] = polyval(vx, points[i, 0:1])[0]
        v[i, 1] = polyval(vy, points[i, 0:1])[0]
        v[i, 2] = polyval(vz, points[i, 0:1])[0]

    return v




class Velocity(PointDataFilter):
    '''Append the dimension-wise or absolute velocity to samples of points
    using a 2D fitted polynomial in a rolling window mode.

    Filter signature:

    ::

        PointData -> Velocity.fit_sample -> PointData

    If Numba is installed, a fast, natively-compiled algorithm is used.

    If `absolute = False`, the "vx", "vy" and "vz" columns are appended. If
    `absolute = True`, then the "v" column is appended.
    '''

    def __init__(self, window, degree = 2, absolute = False):
        if isinstance(nb, NotInstalled):
            warnings.warn((
                "Numba is not installed, so this function will be very slow. "
                "Install Numba to JIT compile the compute-intensive part."
            ), RuntimeWarning)

        self.window = int(window)
        assert self.window % 2 == 1, "The `window` must be an odd number!"
        self.degree = int(degree)
        self.absolute = bool(absolute)

        assert self.window > self.degree, "The `window` must be >`degree`!"


    def fit_sample(self, samples):
        if not isinstance(samples, PointData):
            samples = PointData(samples)

        print(f"len(samples.points) = {len(samples.points)}")

        if not len(samples.points):
            return self._empty_sample(samples)

        if self.window >= len(samples.points):
            return self._invalid_sample(samples)

        vels = compute_velocities(samples.points, self.window, self.degree)

        # Create new object like sample with the extra velocity columns
        if self.absolute:
            absvels = np.linalg.norm(vels, axis = 1)
            points = np.c_[samples.points, absvels]
            columns = samples.columns + ["v"]
        else:
            points = np.c_[samples.points, vels]
            columns = samples.columns + ["vx", "vy", "vz"]

        new_sample = samples.copy(data = points, columns = columns)

        return new_sample


    def _empty_sample(self, samples):
        if self.absolute:
            columns = samples.columns + ["v"]
        else:
            columns = samples.columns + ["vx", "vy", "vz"]

        return samples.copy(
            data = np.empty((0, len(columns))),
            columns = columns,
        )


    def _invalid_sample(self, samples):
        if self.absolute:
            columns = samples.columns + ["v"]
            vels = np.full(len(samples.points), np.nan)
        else:
            columns = samples.columns + ["vx", "vy", "vz"]
            vels = np.full((len(samples.points), 3), np.nan)

        return samples.copy(
            data = np.c_[samples.points, vels],
            columns = columns,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : parallel_screens.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.08.2019


import  time
import  numpy           as      np
from    scipy.integrate import  quad

from    pept            import  LineData
from    pept.base       import  PEPTObject
from    pept.utilities  import  read_csv




def parallel_screens(
    filepath_or_array,
    screen_separation,
    sample_size = None,
    overlap = None,
    verbose = True,
    **kwargs,
):
    '''Initialise PEPT LoRs for parallel screens PET/PEPT detectors from an
    input CSV file or array.

    **The expected data columns in the file are `[time, x1, y1, x2, y2]`**.
    This is automatically transformed into the standard `Lines` format with
    columns being `[time, x1, y1, z1, x2, y2, z2]`, where `z1 = 0` and
    `z2 = screen_separation`.

    `ParallelScreens` can be initialised with a predefined numpy array of LoRs
    or read data from a `.csv`.

    Parameters
    ----------
    filepath_or_array : [str, pathlib.Path, IO] or numpy.ndarray (N, 5)
        A path to a file to be read from or an array for initialisation. A
        path is a string with the (absolute or relative) path to the data
        file or a URL from which the PEPT data will be read. It should
        include the full file name, along with its extension (.csv, .a01,
        etc.).

    screen_separation : float
        The separation (in *mm*) between the two PEPT screens corresponding
        to the `z` coordinate of the second point defining each line. The
        attribute `lines`, with columns
        `[time, x1, y1, z1, x2, y2, z2]`, will have `z1 = 0` and
        `z2 = screen_separation`.

    sample_size : int, default 0
        An `int` that defines the number of lines that should be returned
        when iterating over `lines`. A `sample_size` of 0 yields all the
        data as one single sample. A good starting value would be 200 times
        the maximum number of tracers that would be tracked.

    overlap : int, default 0
        An `int` that defines the overlap between two consecutive samples
        that are returned when iterating over `lines`. An overlap of 0
        implies consecutive samples, while an overlap of
        (`sample_size` - 1) means incrementing the samples by one. A
        negative overlap means skipping values between samples. An error is
        raised if `overlap` is larger than or equal to `sample_size`.

    verbose : bool, default True
        An option that enables printing the time taken for the
        initialisation of an instance of the class. Useful when reading
        large files (10gb files for PEPT data is not unheard of).

    kwargs : other keyword arguments
        Other keyword arguments to be passed to `pept.read_csv`, e.g.
        "skiprows" or "max_rows". See the `pept.read_csv` documentation for
        other arguments.

    Returns
    -------
    LineData
        The initialised LoRs.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size`. Overlap has to be smaller than
        `sample_size`. Note that it can also be negative.

    ValueError
        If the data file does not have the (N, M >= 5) shape.


    Examples
    --------
    Initialise a `LineData` array for three LoRs on a parallel screens
    PEPT scanner (i.e. each line is defined by **two** points each) with a
    head separation of 500 mm:

    >>> lors_raw = np.array([
    >>>     [2, 100, 150, 200, 250],
    >>>     [4, 350, 250, 100, 150],
    >>>     [6, 450, 350, 250, 200]
    >>> ])

    >>> screen_separation = 500
    >>> lors = pept.scanners.parallel_screens(lors_raw, screen_separation)
    Initialised PEPT data in 0.001 s.

    >>> lors
    LineData
    --------
    sample_size = 0
    overlap =     0
    samples =     1
    lines =
      [[  2. 100. 150.   0. 200. 250. 500.]
       [  4. 350. 250.   0. 100. 150. 500.]
       [  6. 450. 350.   0. 250. 200. 500.]]
    lines.shape = (3, 7)
    columns = ['t', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2']


    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    if verbose:
        start = time.time()

    # Check wheter input is a valid `pandas.read_csv` filepath or a numpy
    # array in an "Easier To Ask Forgiveness Than Permission" way.
    try:
        # Try to read the LoR data from `filepath_or_array`.
        # Check if an error is raised when reading file lines using
        lines = read_csv(filepath_or_array, **kwargs)
    except ValueError:
        # Seems like it is an array!
        lines = np.asarray(filepath_or_array, order = "C", dtype = float)

    # Add Z1 and Z2 columns => [time, X1, Y1, Z1, X2, Y2, Z2]
    # Z1 = 0
    lines = np.insert(lines, 3, 0.0, axis = 1)

    # Z2 = `separation`
    lines = np.insert(lines, 6, screen_separation, axis = 1)

    if verbose:
        end = time.time()
        print(f"Initialised PEPT data in {end - start:3.3f} s.\n")

    return LineData(lines, sample_size = sample_size, overlap = overlap)




class ADACGeometricEfficiency(PEPTObject):
    '''Compute the geometric efficiency of a parallel screens PEPT detector at
    different 3D coordinates using Antonio Guida's formula [1]_.

    The default `xlim` and `ylim` values represent the active detector area of
    the ADAC Forte scanner used at the University of Birmingham, but can be
    changed to any parallel screens detector active area range.

    This class assumes PEPT coordinates, with the Y and Z axes being swapped,
    such that Y points upwards and Z is perpendicular to the two detectors.

    Attributes
    ----------
    xlim : (2,) np.ndarray, default [111.78, 491.78]
        The limits of the active detector area in the *x*-dimension.

    ylim : (2,) np.ndarray, default [46.78, 556.78]
        The limits of the active detector area in the *y*-dimension.

    zlim : (2,) np.ndarray
        The limits of the active detector area in the *z*-dimension.

    Examples
    --------
    Simply instantiate the class with the head separation, then 'call' it with
    the (x, y, z) coordinates of the point at which to evaluate the geometric
    efficiency:

    >>> import pept
    >>> separation = 500
    >>> geom = pept.scanners.ADACGeometricEfficiency(separation)
    >>> eg = geom(250, 250, 250)

    Alternatively, the separation may be specified using the both the starting
    and ending limits:

    >>> separation = [-10, 510]
    >>> geom = pept.scanners.ADACGeometricEfficiency(separation)
    >>> eg = geom(250, 250, 250)

    You can evaluate multiple points by using a list / array of values:

    >>> geom([250, 260], 250, 250)
    array([0.18669302, 0.19730517])

    Compute the variation in geometric efficiency in the XY plane:

    >>> separation = 500
    >>> geom = pept.scanners.ADACGeometricEfficiency(separation)

    >>> # Range of x, y values to evaluate the geometric efficiency at
    >>> import numpy as np
    >>> x = np.linspace(120, 480, 100)
    >>> y = np.linspace(50, 550, 100)
    >>> z = 250

    >>> # Evaluate EG on a 2D grid of values at all combinations of x, y
    >>> xx, yy = np.meshgrid(x, y)
    >>> eg = geom(xx, yy, z)

    The geometric efficiencies can be visualised using a Plotly heatmap or
    contour plot:

    >>> import plotly.graph_objs as go
    >>> fig = go.Figure()
    >>> fig.add_trace(go.Contour(x = x, y = y, z = eg))
    >>> fig.show()

    For an interactive 3D volumetric / voxel plot, you can use PyVista:

    >>> # Import necessary libraries; you may need to install PyVista
    >>> import numpy as np
    >>> import pept
    >>> import pyvista as pv

    >>> # Instantiate the ADACGeometricEfficiency class
    >>> geom = pept.scanners.ADACGeometricEfficiency(500)

    >>> # Lower and upper corners of the grid over which to compute the GE
    >>> lower = np.array([115, 50, 5])
    >>> upper = np.array([490, 550, 495])

    >>> # Create 3D meshgrid of values and evaluate the GE at each point
    >>> n = 40
    >>> x = np.linspace(lower[0], upper[0], n)
    >>> y = np.linspace(lower[1], upper[1], n)
    >>> z = np.linspace(lower[2], upper[2], n)
    >>> xx, yy, zz = np.meshgrid(x, y, z)
    >>> eg = geom(xx, yy, zz)

    >>> # Create PyVista grid of values
    >>> grid = pv.UniformGrid()
    >>> grid.dimensions = np.array(eg.shape) + 1
    >>> grid.origin = lower
    >>> grid.spacing = (upper - lower) / n
    >>> grid.cell_arrays["values"] = eg.flatten(order="F")

    >>> # Create PyVista volumetric / voxel plot with an interactive clipper
    >>> p = pv.Plotter()
    >>> p.add_mesh_clip_plane(grid)
    >>> p.show()

    References
    ----------
    .. [1] Guida A. Positron emission particle tracking applied to solid-liquid
       mixing in mechanically agitated vessels (Doctoral dissertation,
       University of Birmingham).

    '''

    def __init__(
        self,
        separation,
        xlim = [111.78, 491.78],
        ylim = [46.78, 556.78],
    ):
        # Separation may be either a number (e.g. 500) or a 2-list ([-10, 510])
        separation = np.array(separation)
        if separation.shape == ():
            self.zlim = np.array([0, separation])
        else:
            self.zlim = separation

        self.xlim = np.array(xlim)
        self.ylim = np.array(ylim)

        self.veg = np.vectorize(self.eg, cache = True)


    def eg(self, x, y, z):
        '''Return the geometric efficiency evaluated at a single point
        (x, y, z) *in PEPT coordinates*, i.e. Y points upwards.
        '''

        # Translate x, y, z relative to the scanner active area's centre
        x = (x - self.xlim[0]) - (self.xlim[1] - self.xlim[0]) / 2
        y = (y - self.ylim[0]) - (self.ylim[1] - self.ylim[0]) / 2
        z = (z - self.zlim[0]) - (self.zlim[1] - self.zlim[0]) / 2

        # Active detector area dimensions
        ld = self.xlim[1] - self.xlim[0]
        hd = self.ylim[1] - self.ylim[0]
        sd = self.zlim[1] - self.zlim[0]

        # Solid horizontal (XZ in PEPT coordinates) angle
        theta_min = np.arctan(max(
            (sd - 2 * z) / (ld - 2 * x),
            (sd + 2 * z) / (ld + 2 * x),
        ))

        theta_max = np.pi - np.arctan(max(
            (sd - 2 * z) / (ld + 2 * x),
            (sd + 2 * z) / (ld - 2 * x),
        ))

        # Terms for the two integrals that will be summed
        m1 = max(
            (sd - 2 * z) / (hd - 2 * y),
            (sd + 2 * z) / (hd + 2 * y),
        )

        m2 = max(
            (sd - 2 * z) / (hd + 2 * y),
            (sd + 2 * z) / (hd - 2 * y),
        )

        # Evaluate integrals
        f1_integral, _ = quad(
            lambda theta: (m1**2 / np.sin(theta)**2 + 1)**(-0.5),
            theta_min, theta_max
        )

        f2_integral, _ = quad(
            lambda theta: (m2**2 / np.sin(theta)**2 + 1)**(-0.5),
            theta_min, theta_max
        )

        return (f1_integral + f2_integral) / (2 * np.pi)


    def __call__(self, x, y, z):
        # Allow 'calling' the class using a vectorized call to `ge`
        return self.veg(x, y, z)


    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"xlim={self.xlim}, "
            f"ylim={self.ylim}, "
            f"zlim={self.zlim})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : fpi.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 16.04.2021


import  warnings

import  numpy               as      np

from    beartype            import  beartype

import  pept
from    .fpi_ext            import  fpi_ext


class FPI(pept.base.VoxelsFilter):
    '''FPI is a modern voxel-based tracer-location algorithm that can reliably
    work with unknown numbers of tracers in fast and noisy environments.

    It was successfully used to track fast-moving radioactive tracers in pipe
    flows at the Virginia Commonwealth University. If you use this algorithm in
    your work, please cite the following paper:

        Wiggins C, Santos R, Ruggles A. A feature point identification method
        for positron emission particle tracking with multiple tracers. Nuclear
        Instruments and Methods in Physics Research Section A: Accelerators,
        Spectrometers, Detectors and Associated Equipment. 2017 Jan 21;
        843:22-8.

    Permission was granted by Dr. Cody Wiggins in March 2021 to publish his
    code in the `pept` library under the GNU v3.0 license.

    Two main methods are provided: `fit_sample` for tracking a single voxel
    space (i.e. a single `pept.Voxels`) and `fit` which tracks all the samples
    encapsulated in a `pept.VoxelData` class *in parallel*.

    Attributes
    ----------
    w: double
        Search range to be used in local maxima calculation. Typical values for
        w are 2 - 5 (lower number for more particles or smaller particle
        separation).

    r: double
        Fraction of peak value used as threshold. Typical values for r are
        usually between 0.3 and 0.6 (lower for more particles, higher for
        greater background noise)

    lld_counts: double, default 0
        A secondary lld to prevent assigning local maxima to voxels with very
        low values. The parameter lld_counts is not used much in practice -
        for most cases, it can be set to zero.

    Examples
    --------
    A typical workflow would involve reading LoRs from a file, creating a lazy
    `VoxelData` voxellised representation, instantiating an `FPI` class,
    tracking the tracer locations from the LoRs, and plotting them.

    >>> import pept
    >>>
    >>> lors = pept.LineData(...)   # set sample_size and overlap appropriately
    >>> voxels = pept.tracking.Voxelize((50, 50, 50)).fit(lors)
    >>>
    >>> fpi = pept.tracking.FPI(w = 3, r = 0.4)
    >>> positions = fpi.fit(voxels) # this is a `pept.PointData` instance

    A much more efficient approach would be to create a `pept.Pipeline`
    containing a voxelization step and then FPI:

    >>> from pept.tracking import *
    >>>
    >>> pipeline = Voxelize((50, 50, 50)) + FPI() + Stack()
    >>> positions = pipeline.fit(lors)

    Finally, plotting results:

    >>> from pept.plots import PlotlyGrapher
    >>>
    >>> grapher = PlotlyGrapher()
    >>> grapher.add_points(positions)
    >>> grapher.show()

    >>> from pept.plots import PlotlyGrapher2D
    >>> PlotlyGrapher2D().add_timeseries(positions).show()

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.

    '''

    def __init__(
        self,
        w = 3.,
        r = 0.4,
        lld_counts = 0.,
        verbose = False,
    ):
        '''`FPI` class constructor.

        Parameters
        ----------
        w: double
            Search range to be used in local maxima calculation. Typical values
            for w are 2 - 5 (lower number for more particles or smaller
            particle separation).

        r: double
            Fraction of peak value used as threshold. Typical values for r are
            usually between 0.3 and 0.6 (lower for more particles, higher for
            greater background noise)

        lld_counts: double, default 0
            A secondary lld to prevent assigning local maxima to voxels with
            very low values. The parameter `lld_counts` is not used much in
            practice - for most cases, it can be set to zero.

        verbose: bool, default False
            Show extra information on class instantiation.

        '''
        self.w = float(w)
        self.r = float(r)
        self.lld_counts = float(lld_counts)


    @beartype
    def fit_sample(self, voxels: pept.Voxels):
        '''Use the FPI algorithm to locate a tracer from a single voxellised
        space (i.e. from one sample of LoRs).

        A sample of LoRs can be voxellised using the `pept.Voxels.from_lines`
        method before calling this function.

        Parameters
        ----------
        voxels: pept.Voxels
            A single voxellised space (i.e. from a single sample of LoRs) for
            which the tracers' locations will be found using the FPI method.

        timestamp: float, default 0.
            The timestamp to associate with the tracer positions found in this
            voxel space.

        as_array: bool, default False
            If `True`, return the found tracers' locations as a NumPy array.
            Otherwise, return them in a `pept.PointData` instance.

        verbose: bool, default False
            Show extra information on the sample processing step.

        Returns
        -------
        locations: numpy.ndarray or pept.PointData
            The tracked locations found; if `as_array` is True, they are
            returned as a NumPy array with columns [time, x, y, z, error_x,
            error_y, error_z]. If `as_array` is False, the points are returned
            in a `pept.PointData` for ease of visualisation.

        Raises
        ------
        TypeError
            If `voxels` is not an instance of `pept.Voxels` (or subclass
            thereof).
        '''

        points = fpi_ext(
            voxels.voxels,
            self.w,
            self.r,
            self.lld_counts,
        )

        # Insert the time column and translate the coordinates from the voxel
        # space to the physical space
        points[:, 1:4] *= voxels.voxel_size
        points[:, 1:4] += [voxels.xlim[0], voxels.ylim[0], voxels.zlim[0]]

        points = np.insert(points, 0, np.nan, axis = 1)
        if "_lines" in voxels.attrs:
            points[:, 0] = voxels.attrs["_lines"].lines[:, 0].mean()
        else:
            warnings.warn((
                "The input `Voxels` did not have a '_lines' attribute, so no "
                "timestamp can be inferred. The time was set to NaN."
            ), RuntimeWarning)

        return pept.PointData(
            points,
            columns = ["t", "x", "y", "z", "error_x", "error_y", "error_z"],
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : transformers.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 08.08.2021


import  re
import  sys
import  warnings
from    typing          import  Union
from    numbers         import  Number

if sys.version_info.minor >= 9:
    # Python 3.9
    from collections.abc import  Iterable
else:
    from typing         import  Iterable

import  textwrap

import  numpy           as      np

from    beartype        import  beartype

from    pept.base       import  PointData, LineData
from    pept.base       import  Filter, Reducer, IterableSamples


PointsOrLines = Union[PointData, LineData]




class Stack(Reducer):
    '''Stack iterables - for example a ``list[pept.LineData]`` into a single
    ``pept.LineData``, a ``list[list]`` into a flattened ``list``.

    Reducer signature:

    ::

        list[LineData] -> Stack.fit -> LineData
        list[PointData] -> Stack.fit -> PointData

        list[list[Any]] -> Stack.fit -> list[Any]
        list[numpy.ndarray] -> Stack.fit -> numpy.ndarray

        other -> Stack.fit -> other

    Can optionally set a given `sample_size` and `overlap`. This is useful
    when collecting a list of processed samples back into a single object.
    '''

    def __init__(self, sample_size = None, overlap = None):
        self.sample_size = sample_size
        self.overlap = overlap


    @beartype
    def fit(self, samples: Iterable):
        # If it's a LineData / PointData, the `samples` are already stacked.
        # Simply set the sample_size and overlap if required and return them
        if isinstance(samples, IterableSamples):
            if self.sample_size is not None:
                samples.sample_size = self.sample_size
            if self.sample_size is not None:
                samples.overlap = self.overlap
            return samples

        # If it's an empty iterator, we don't have anything to stack
        if len(samples) == 0:
            return samples

        # Stack Lines into LineData
        elif isinstance(samples[0], LineData):
            samples = LineData(samples)

        # Stack Points into PointData
        elif isinstance(samples[0], PointData):
            samples = PointData(samples)

        # Flatten list of lists
        elif isinstance(samples[0], list):
            samples = [item for sublist in samples for item in sublist]

        # Vertically stack list of NumPy arrays
        elif isinstance(samples[0], np.ndarray):
            samples = np.vstack(samples)

        # Set new sample_size and overlap if required
        if self.sample_size is not None:
            samples.sample_size = self.sample_size
        if self.sample_size is not None:
            samples.overlap = self.overlap

        return samples




class SplitLabels(Filter):
    '''Split a sample of data into unique ``label`` values, optionally removing
    noise and extracting `_lines` attributes.

    Filter signature:

    ::

        # `extract_lines` = False (default)
        LineData -> SplitLabels.fit_sample -> list[LineData]
        PointData -> SplitLabels.fit_sample -> list[PointData]

        # `extract_lines` = True and PointData.lines exists
        PointData -> SplitLabels.fit_sample -> list[LineData]


    The sample of data must have a column named exactly "label". The filter
    normally removes the "label" column in the output (if
    ``remove_label = True``).
    '''

    def __init__(
        self,
        remove_labels = True,
        extract_lines = False,
        noise = False,
    ):
        self.remove_labels = bool(remove_labels)
        self.extract_lines = bool(extract_lines)
        self.noise = bool(noise)


    def _get_cluster(self, sample, labels_mask, lines_cols = None):
        # Extract the labels column
        cluster_data = sample.data[labels_mask]

        if lines_cols is not None:
            line_indices = np.unique(cluster_data[:, lines_cols])
            lines = sample.attrs["_lines"].lines
            cluster_lines = lines[line_indices.astype(int)]

        if self.extract_lines:
            return sample.attrs["_lines"].copy(data = cluster_lines)

        cluster = sample.copy(data = cluster_data)
        if lines_cols is not None:
            cluster.attrs["_lines"] = sample.attrs["_lines"].copy(
                data = cluster_lines,
            )

        return cluster


    def _empty_cluster(self, sample, lines_cols = None):
        if self.extract_lines:
            # Return empty LineData
            return sample.attrs["_lines"].copy(
                data = sample.attrs["_lines"][0:0],
            )

        cluster = sample.copy(data = sample[0:0])
        if lines_cols is not None:
            cluster.attrs["_lines"] = sample.attrs["_lines"].copy(
                data = sample.attrs["_lines"][0:0],
            )

        return cluster


    @beartype
    def fit_sample(self, sample: IterableSamples):
        # Extract the labels column
        col_idx = sample.columns.index("label")
        labels = sample.data[:, col_idx]

        # Check if there is a `._lines` attribute with `line_index` columns
        lines_cols = None
        if "_lines" in sample.attrs:
            lines_cols = [
                i for i, c in enumerate(sample.columns)
                if c.startswith("line_index")
            ]

            if len(lines_cols) == 0:
                warnings.warn((
                    "A `_lines` attribute was found, but no lines can "
                    "be extracted without columns `line_index<N>`."
                ), RuntimeWarning)

                lines_cols = None
                self.extract_lines = False

        elif self.extract_lines:
            raise ValueError(textwrap.fill((
                "If `extract_lines` is True, then the input `sample` must "
                "contain a `_lines` attribute."
            )))

        # If noise is requested, also include the noise cluster
        if self.noise:
            labels_unique = np.unique(labels)
        else:
            labels_unique = np.unique(labels[labels != -1])

        # For each unique label, create a new PointData / LineData cluster that
        # maintains / propagates all attributes (which needs a copy)
        clusters = [
            self._get_cluster(sample, labels == label, lines_cols)
            for label in labels_unique
        ]

        # If no valid cluster was found, return at least a single empty cluster
        if not len(clusters):
            clusters.append(self._empty_cluster(sample, lines_cols))

        # Remove the "label" column if needed
        if self.remove_labels and not self.extract_lines:
            for i in range(len(clusters)):
                clusters[i] = clusters[i].copy(
                    data = np.delete(clusters[i].data, col_idx, axis = 1),
                    columns = (clusters[i].columns[:col_idx] +
                               clusters[i].columns[col_idx + 1:]),
                )

        return clusters




class Centroids(Filter):
    '''Compute the geometric centroids of a list of samples of points.

    Filter signature:

    ::

        PointData -> Centroids.fit_sample -> PointData
        list[PointData] -> Centroids.fit_sample -> PointData
        numpy.ndarray -> Centroids.fit_sample -> PointData

    This filter can be used right after ``pept.tracking.SplitLabels``, e.g.:

    >>> (SplitLabels() + Centroids()).fit(points)

    '''


    def __init__(self, error = False, cluster_size = False):
        self.error = bool(error)
        self.cluster_size = bool(cluster_size)


    def _empty_centroid(self, points):
        # Return an empty centroid with the correct number of columns
        ncols = points.shape[1]
        if self.error:
            ncols += 1
        if self.cluster_size:
            ncols += 1
        return np.empty((0, ncols))


    def _centroid(self, points):
        if len(points) == 0:
            return self._empty_centroid(points)

        c = points.mean(axis = 0)

        # If error is requested, compute std-dev of distances from centroid
        if self.error:
            err = np.linalg.norm(points - c, axis = 1).std()
            c = np.r_[c, err]

        # If cluster_size is requested, also append the number of points
        if self.cluster_size:
            c = np.r_[c, len(points)]

        return c


    def fit_sample(self, points):
        # Type-checking inputs
        if isinstance(points, PointData):
            list_points = [points]
        elif isinstance(points, np.ndarray):
            list_points = [PointData(points)]
        else:
            list_points = list(points)

        # Compute centroid for each PointData and stack centroid arrays
        centroids = np.vstack([self._centroid(p.points) for p in list_points])
        attributes = list_points[0].extra_attrs()

        # If error or cluster_size are requested, append those columns
        columns = list_points[0].columns

        if self.error:
            columns.append("error")

        if self.cluster_size:
            columns.append("cluster_size")

        return PointData(centroids, columns = columns, **attributes)




class LinesCentroids(Filter):
    '''Compute the minimum distance point of some ``pept.LineData`` while
    iteratively removing a fraction of the furthest lines.

    Filter signature:

    ::

        list[LineData] -> LinesCentroids.fit_sample -> PointData
        LineData -> LinesCentroids.fit_sample -> PointData
        numpy.ndarray -> LinesCentroids.fit_sample -> PointData

    The code below is adapted from the PEPT-EM algorithm developed by Antoine
    Renaud and Sam Manger
    '''

    def __init__(self, remove = 0.1, iterations = 6):
        self.remove = float(remove)
        self.iterations = int(iterations)


    @staticmethod
    def centroid(lors):
        nx = np.newaxis

        m = np.identity(3)[nx, :, :] - lors[:, nx, 4:7] * lors[:, 4:7, nx]
        n = np.sum(m, axis = 0)
        v = np.sum(np.sum(m * lors[:, nx, 1:4], axis=-1), axis=0)

        return np.matmul(np.linalg.inv(n), v)


    @staticmethod
    def distance_matrix(x, lors):
        y = x[np.newaxis, :3] - lors[:, 1:4]
        return np.sum(y**2, axis=-1) - np.sum(y * lors[:, 4:7], axis=-1)**2


    def predict(self, lines):
        # Rewrite LoRs in the vectorial form y(x) = position + x * direction
        lors = lines[:, :7].copy(order = "C")

        lors[:, 4:7] = lors[:, 4:7] - lors[:, 1:4]
        lors[:, 4:7] /= np.linalg.norm(lors[:, 3:], axis = -1)[:, np.newaxis]

        # Begin with equal weights for all LoRs
        weights = np.ones(len(lors))
        x = LinesCentroids.centroid(lors)

        # Iteratively remove the furthest LoRs and recompute centroid
        for i in range(self.iterations):
            d2 = LinesCentroids.distance_matrix(x, lors)

            k = int(len(d2) * (1 - self.remove * (i + 1)))
            part = np.argpartition(d2, k)
            weights[part[k:]] = 0

            x = LinesCentroids.centroid(lors)

        # Add timestamp as the mean LoRs' time
        return np.hstack((lors[:, 0].mean(), x))


    def fit_sample(self, lines):
        # Type-checking inputs
        if isinstance(lines, LineData):
            list_lines = [lines]
        elif isinstance(lines, np.ndarray):
            list_lines = [LineData(lines)]
        else:
            list_lines = list(lines)

        centroids = [self.predict(lines.lines) for lines in list_lines]
        return PointData(np.vstack(centroids), **list_lines[0].extra_attrs())




class Condition(Filter):
    '''Select only data satisfying multiple conditions, given as a string, a
    function or list thereof; e.g. ``Condition("error < 15")`` selects all
    points whose "error" column value is smaller than 15.

    Filter signature:

    ::

        PointData -> Condition.fit_sample -> PointData
        LineData -> Condition.fit_sample -> LineData

    In the simplest case, a column name is specified, plus a comparison, e.g.
    ``Condition("error < 15, y > 100")``; multiple conditions may be
    concatenated using a comma.

    More complex conditions - where the column name is not the first operand -
    can be constructed using single quotes, e.g. using NumPy functions in
    ``Condition("np.isfinite('x')")`` to filter out NaNs and Infs. Quotes can
    be used to index columns too: ``Condition("'0' < 150")`` selects all rows
    whose first column is smaller than 150.

    Generally, you can use any function returning a boolean mask, either as a
    string of code ``Condition("np.isclose('x', 3)")`` or a user-defined
    function receiving a NumPy array ``Condition(lambda x: x[:, 0] < 10)``.

    Finally, multiple such conditions may be supplied separately:
    ``Condition(lambda x: x[:, -1] > 10, "'t' < 50")``.
    '''

    def __init__(self, *conditions):
        # Calls the conditions setter which does parsing
        self.conditions = conditions


    @property
    def conditions(self):
        return self._conditions


    @conditions.setter
    def conditions(self, conditions):
        if isinstance(conditions, str):
            self._conditions = Condition._parse_condition(conditions)
        elif callable(conditions):
            self._conditions = [conditions]
        else:
            cs = []
            for cond in conditions:
                cs.extend(Condition._parse_condition(cond))
            self._conditions = cs


    @staticmethod
    def _parse_condition(cond):
        if callable(cond):
            return [cond]

        conditions = str(cond).replace(" ", "").split(",")

        # Compile regex object to find quoted strings
        finder = re.compile(r"'\w+'")

        for i in range(len(conditions)):
            # Replace single-quoted column numbers / names
            if "'" in conditions[i]:
                conditions[i] = finder.sub(
                    Condition._replace_quoted,
                    conditions[i],
                )
                continue

            # If condition is a simple comparison, allow using non-quoted
            # column names
            op = None
            if "<" in conditions[i]:
                op = "<"
            elif ">" in conditions[i]:
                op = ">"
            elif "!" in conditions[i]:
                op = "!"
            elif "==" in conditions[i]:
                op = "=="

            if op is not None:
                cs = conditions[i].split(op)
                cs[0] = Condition._replace_term(cs[0])
                conditions[i] = op.join(cs)

            else:
                raise ValueError(textwrap.fill((
                    f"The input `conditions[i] = {conditions[i]}` did not "
                    "contain an operator or single-quoted terms."
                )))

        return conditions


    @staticmethod
    def _replace_term(term: str):
        return f"data[:, sample.columns.index('{term}')]"


    @staticmethod
    def _replace_quoted(term):
        # Remove single quotes
        if isinstance(term, re.Match):
            term = term.group()
        term = term.split("'")[1]

        try:
            index = int(term)
            return f"data[:, {index}]"
        except ValueError:
            return f"data[:, sample.columns.index('{term}')]"


    @beartype
    def fit_sample(self, sample: IterableSamples):
        data = sample.data

        for cond in self.conditions:
            if callable(cond):
                data = data[cond(data)]
            else:
                data = data[eval(cond, globals(), locals())]

        return sample.copy(data = data)




class Remove(Filter):
    '''Remove columns (either column names or indices) from `pept.LineData` or
    `pept.PointData`.

    Filter signature:

    ::

        pept.LineData  -> Remove.fit_sample -> pept.LineData
        pept.PointData -> Remove.fit_sample -> pept.PointData

    Examples
    --------
    To remove a single column named "line_index":

    >>> import pept
    >>> from pept.tracking import *
    >>> points = pept.PointData(...)    # Some dummy data

    >>> rem = Remove("line_index")
    >>> points_without = rem.fit_sample(points)

    Remove all columns starting with "line_index" using a glob operator (*):

    >>> points_without = Remove("line_index*").fit_sample(points)

    Remove the first column based on its index:

    >>> points_without = Remove(0).fit_sample(points)

    Finally, multiple removals may be chained into a list:

    >>> points_without = Remove(["line_index*", -1]).fit_sample(points)

    '''

    def __init__(self, *columns):
        self._indices = []
        self._filters = []

        # Calls the `columns` setter which does parsing
        self.columns = columns


    @property
    def columns(self):
        return self._columns


    @columns.setter
    def columns(self, columns):
        self._columns = [Remove._parse(col) for col in columns]

        # Split the removers into regex strings and column indices
        for c in self._columns:
            if isinstance(c, str):
                self._filters.append(c)
            else:
                self._indices.append(c)


    @staticmethod
    def _parse(col):
        if isinstance(col, str):
            return col.replace("*", r"\w*")
        elif isinstance(col, Number):
            return int(col)
        else:
            raise ValueError(textwrap.fill((
                "Each input argument in `columns` must be a string or an "
                f"integer. One of them was `type(col) = {type(col)}`."
            )))


    @beartype
    def fit_sample(self, sample: IterableSamples):
        # Extract the relevant `sample` attributes
        columns = sample.columns
        ncols = len(columns)

        # The regex filters to use and column numbers to remove
        filters = self._filters
        indices = self._indices

        # Column indices to remove and remaining column names
        removed = set()
        columns_filtered = []

        for i, c in enumerate(columns):
            # Also handle negative indices
            if any((re.fullmatch(r, c) for r in filters)) or \
                    any((i == ind or i == ind + ncols for ind in indices)):
                removed.add(i)
            else:
                columns_filtered.append(c)

        indices_filtered = [i for i in range(len(columns)) if i not in removed]
        data = sample.data[:, indices_filtered]

        return sample.copy(data = data, columns = columns_filtered)




class SplitAll(Reducer):
    '''Stack all samples and split them into a list according to a named /
    numeric column index.

    Reducer signature:

    ::

        LineData -> SplitAll.fit -> list[LineData]
        list[LineData] -> SplitAll.fit -> list[LineData]

        PointData -> SplitAll.fit -> list[PointData]
        list[PointData] -> SplitAll.fit -> list[PointData]

        numpy.ndarray -> SplitAll.fit -> list[numpy.ndarray]
        list[numpy.ndarray] -> SplitAll.fit -> list[numpy.ndarray]

    If using a LineData / PointData, you can use a columns name as a string,
    e.g. ``SplitAll("label")`` or a number ``SplitAll(4)``. If using a NumPy
    array, only numeric indices are accepted.
    '''

    def __init__(self, column):
        try:
            self.column_index = int(column)
            self.column_name = None
        except ValueError:
            self.column_name = str(column)
            self.column_index = None


    @beartype
    def fit(self, samples: Iterable):
        # Reduce / stack list of samples onto a single IterableSamples / array
        samples = Stack().fit(samples)

        if isinstance(samples, np.ndarray):
            return self._split_numpy(samples)
        elif isinstance(samples, IterableSamples):
            return self._split_iterable_samples(samples)
        else:
            raise TypeError(textwrap.fill((
                "The input samples must be NumPy arrays, PointData / LineData "
                f"or lists thereof. Received `type(samples) = {type(samples)}`"
            )))


    def _split_numpy(self, samples):
        if self.column_index is None:
            raise TypeError(textwrap.fill((
                "If the samples are NumPy arrays, you must use a numeric "
                f"column index; used a named column: `{self.column_name}`."
            )))

        col_data = samples[:, self.column_index]
        labels = np.unique(col_data)

        # If no labels exist, return a list with an empty sample
        if not len(labels):
            return [samples[0:0]]

        return [samples[col_data == label] for label in labels]


    def _split_iterable_samples(self, samples):
        if self.column_index is not None:
            col_data = samples.data[:, self.column_index]
        else:
            col_data = samples.data[:, samples.columns.index(self.column_name)]

        labels = np.unique(col_data)

        # If no labels exist, return a list with an empty sample
        if not len(labels):
            return [samples[0:0]]

        return [
            samples.copy(data = samples.data[col_data == label])
            for label in labels
        ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you should cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : iterable_samples.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 09.04.2020


import  pickle
import  operator
import  textwrap
from    copy                import  copy as shallowcopy
from    copy                import  deepcopy
from    dataclasses         import  dataclass
from    numbers             import  Number
from    collections.abc     import  Collection
from    concurrent.futures  import  ThreadPoolExecutor

import  numpy               as      np

from    tqdm                import  tqdm




class PEPTObject:
    '''Base class for all PEPT-oriented objects.'''

    def copy(self, deep = True):
        '''Create a deep copy of an instance of this class, including all
        inner attributes.
        '''
        return deepcopy(self) if deep else shallowcopy(self)


    def save(self, filepath):
        '''Save a `PEPTObject` instance as a binary `pickle` object.

        Saves the full object state, including inner attributes, in a
        portable binary format. Load back the object using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `LineData` instance, then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> lines.save("lines.pickle")

        >>> lines_reloaded = pept.LineData.load("lines.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `PEPTObject` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.PEPTObject subclass instance
            The loaded object.

        Examples
        --------
        Save a `LineData` instance, then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> lines.save("lines.pickle")

        >>> lines_reloaded = pept.LineData.load("lines.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def __repr__(self):
        # Return pretty string representation of an arbitrary object
        docs = []
        for attr in dir(self):
            if not attr.startswith("_"):
                memb = getattr(self, attr)
                if not callable(memb):
                    docs.append(f"{attr} = {memb}")

        name = self.__class__.__name__
        underline = "-" * len(name)
        return f"{name}\n{underline}\n" + "\n".join(docs)




def samples_indices_number(data, sample_size, overlap):
    '''Compute the sample indices given some (N, M) `data` and fixed (integer)
    `sample_size` and `overlap`.

    The samples indices are returned in a (N, 2) NumPy array, where the first
    column is each sample's starting index (in `data`), and the second column
    contains the corresponding sample's ending index, so that e.g. sample `n`
    can be extracted as `data[samples_indices[n, 0]:samples_indices[n, 1]]`.
    '''

    if sample_size == 0:
        return np.zeros((0, 2))

    elif sample_size < 0:
        raise ValueError((
            f"\n[ERROR]: `sample_size = {sample_size}` must be positive "
            "(>= 0).\n"
        ))

    elif overlap >= sample_size:
        raise ValueError((
            f"\n[ERROR]: `overlap = {overlap}` must be smaller than "
            f"`sample_size = {sample_size}`.\n"
        ))

    # The first column is each sample's starting index; the second column is
    # the corresponding sample's ending index
    start = np.arange(0, len(data) - sample_size + 1, sample_size - overlap)
    end = start + sample_size

    return np.c_[start, end]




def samples_indices_iterable(data, sample_size):
    '''Compute the sample indices given each sample's length in an iterable
    `sample_size`.
    '''
    end = np.cumsum(sample_size, dtype = int)
    start = end - sample_size

    return np.c_[start, end]




@dataclass
class TimeWindow:
    '''Define a `sample_size` as a fixed time window / slice. You can use this
    as a direct replacement of the `sample_size` and `overlap`:

    ::

        points = pept.PointData(sample_size = pept.TimeWindow(5.5))

    '''
    __slots__ = ["window"]      # Only have the `window` attribute
    window: float




def samples_indices_time_window(
    data,
    sample_size: TimeWindow,
    overlap: TimeWindow,
):
    '''Compute the sample indices given a time window across the timestamps in
    `data` (i.e. column 0).
    '''

    if sample_size.window == 0:
        return np.zeros((0, 2))

    elif sample_size.window < 0:
        raise ValueError((
            f"\n[ERROR]: `sample_size.window = {sample_size}` must be "
            "positive (>= 0).\n"
        ))

    elif overlap.window >= sample_size.window:
        raise ValueError((
            f"\n[ERROR]: `overlap = {overlap}` must be smaller than "
            f"`sample_size = {sample_size}`.\n"
        ))

    eps = np.finfo(float).resolution

    start_times = np.arange(
        0.9999 * data[0, 0] - eps,
        1.0001 * (data[-1, 0] - sample_size.window) + eps,
        sample_size.window - overlap.window,
        dtype = float,
    )
    end_times = start_times + sample_size.window

    start_indices = np.searchsorted(data[:, 0], start_times, "right")
    end_indices = np.searchsorted(data[:, 0], end_times, "right")

    # Remove empty samples
    cond = start_indices != end_indices
    start_indices = start_indices[cond]
    end_indices = end_indices[cond]

    return np.c_[start_indices, end_indices]




class IterableSamples(PEPTObject, Collection):
    '''An class for iterating through an array (or array-like) in samples with
    potential overlap.

    This class can be used to access samples of data of an adaptive
    ``sample_size`` and ``overlap`` without requiring additional storage.

    The samples from the underlying data can be accessed using both indexing
    (``samples[0]``) and iteration (``for sample in samples: ...``).

    Particular cases:
        1. If sample_size == 0, all data_samples is returned as one single
           sample.
        2. If overlap >= sample_size, an error is raised.
        3. If overlap < 0, lines are skipped between samples.

    Attributes
    ----------
    data : iterable that supports slicing
        An iterable (e.g. numpy array) that supports slicing syntax (data[5:7])
        storing the data that will be iterated over in samples.

    sample_size : int
        The number of rows in `data` to be returned in a single sample. A
        `sample_size` of 0 yields all the data as a single sample.

    overlap : int
        The number of overlapping rows from `data` between two consecutive
        samples. An overlap of 0 implies consecutive samples, while an
        overlap of (`sample_size` - 1) means incrementing the samples by
        one. A negative overlap implies skipping values between samples.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size` unless `sample_size` is 0. Overlap
        must be smaller than `sample_size`. Note that it can also be negative.

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.

    '''

    def __init__(
        self,
        data,
        sample_size = None,
        overlap = None,
        columns = [],
        **kwargs,
    ):
        '''`IterableSamples` class constructor.

        Parameters
        ----------
        data : iterable
            The data that will be iterated over in samples; most commonly a
            NumPy array.

        sample_size : int or Iterable[Int], optional
            The number of rows in `data` to be returned in a single sample. A
            `sample_size` of 0 yields all the data as a single sample.

        overlap : int, optional
            The number of overlapping rows from `data` between two consecutive
            samples. An overlap of 0 implies consecutive samples, while an
            overlap of (`sample_size` - 1) means incrementing the samples by
            one. A negative overlap implies skipping values between samples.

        '''

        # If not enough columns were supplied, append "col4", "col5", etc.
        if len(columns) < data.shape[1]:
            columns = list(columns) + [
                f"col{i}" for i in range(len(columns), data.shape[1])
            ]
        elif len(columns) > data.shape[1]:
            columns = columns[:data.shape[1]]

        self._data = np.asarray(data, dtype = float, order = "C")
        self._columns = columns
        self._attrs = kwargs

        self._set(sample_size = sample_size, overlap = overlap)
        self._index = 0


    def _set(self, sample_size = None, overlap = None):
        # If the overlap is defined, ensure it has the same type as sample_size
        if overlap is not None and not isinstance(overlap, type(sample_size)):
            raise TypeError(textwrap.fill((
                "The input `overlap` (if defined) must have the same type "
                f"as `sample_size`. Received `{type(overlap)}`."
            )))

        # Set sample_size. This calls the setter which does type-checking
        self._overlap = overlap
        self.sample_size = sample_size


    @property
    def data(self):
        return self._data


    @property
    def columns(self):
        return self._columns


    @property
    def attrs(self):
        return self._attrs


    def extra_attrs(self):
        return {k: v for k, v in self.attrs.items() if not k.startswith("_")}


    def hidden_attrs(self):
        return {k: v for k, v in self.attrs.items() if k.startswith("_")}


    @property
    def samples_indices(self):
        return self._samples_indices


    @samples_indices.setter
    def samples_indices(self, samples_indices):
        samples_indices = np.asarray(samples_indices, order = "C", dtype = int)

        if samples_indices.ndim != 2 or samples_indices.shape[1] != 2:
            raise ValueError(textwrap.fill((
                "The `samples_indices`, if given as a NumPy array, must be "
                "a (N, 2) matrix where the first column contains each "
                "sample's starting index (in `data`), and the second "
                "column is the corresponding sample's end index. "
                f"Received array with shape `{samples_indices.shape}`."
            )))

        self._sample_size = None
        self._overlap = None
        self._samples_indices = samples_indices


    @property
    def sample_size(self):
        return self._sample_size


    @sample_size.setter
    def sample_size(self, sample_size):
        if sample_size is None:
            self._sample_size = len(self.data)
            self._overlap = 0
            self._samples_indices = np.array([[0, len(self.data)]])
        elif isinstance(sample_size, Number):
            # If the overlap is of a different type, reset it
            if not isinstance(self.overlap, Number):
                self._overlap = 0

            self._sample_size = int(sample_size)
            self._samples_indices = samples_indices_number(
                self.data, self._sample_size, self._overlap
            )
        elif isinstance(sample_size, TimeWindow):
            # If the overlap is of a different type, reset it
            if not isinstance(self.overlap, TimeWindow):
                self._overlap = TimeWindow(0.)

            self._sample_size = sample_size
            self._samples_indices = samples_indices_time_window(
                self.data, self._sample_size, self._overlap
            )
        elif hasattr(sample_size, "__iter__"):
            sample_size = np.asarray(sample_size, dtype = int)

            # Special case: if all sample_sizes are equal, set them to that
            if len(sample_size) and (sample_size == sample_size[0]).all():
                self._overlap = 0
                self.sample_size = sample_size[0]
                return

            self._overlap = None
            self._sample_size = sample_size

            self._samples_indices = samples_indices_iterable(
                self.data, self._sample_size
            )
        else:
            raise TypeError("The input `sample_size` has an unknown type.")


    @property
    def overlap(self):
        return self._overlap


    @overlap.setter
    def overlap(self, overlap):
        if overlap is not None and not \
                isinstance(overlap, type(self.sample_size)):
            raise TypeError(textwrap.fill((
                "The input `overlap` must have the same type "
                f"as `sample_size`. Received `{type(overlap)}`."
            )))

        # Call the `sample_size` setter which does type checking
        self._overlap = overlap
        self.sample_size = self._sample_size


    def copy(
        self,
        deep = True,
        data = None,
        extra = True,
        hidden = True,
        **attrs,
    ):
        '''Construct a similar object, optionally with different `data`. If
        `extra`, extra attributes are propagated; same for `hidden`.
        '''

        # Propagate needed attributes
        if extra and hidden:
            attrs.update(self.attrs)
        elif extra:
            attrs.update(self.extra_attrs())
        elif hidden:
            attrs.update(self.hidden_attrs())

        set_samples_indices = False
        if data is None:
            data = self.data.copy() if deep else self.data
            attrs["columns"] = self.columns

            # If copying the existing inner data and we have custom
            # samples_indices, set them afterwards
            if self.sample_size is None:
                set_samples_indices = True
            else:
                attrs["sample_size"] = self.sample_size
                attrs["overlap"] = self.overlap

        if "columns" not in attrs:
            attrs["columns"] = self.columns

        new_instance = self.__class__(data, **attrs)

        if set_samples_indices:
            new_instance.samples_indices = self.samples_indices
        return new_instance


    def __len__(self):
        # Defined so that len(class_instance) returns the number of samples.
        return len(self.samples_indices)


    def __contains__(self, key):
        return self.data.__contains__(key)


    def __getitem__(self, n):
        # String indexing into columns
        if isinstance(n, str):
            return self.data[:, self.columns.index(n)]

        # Numeric indexing into samples
        indices = self.samples_indices

        if isinstance(n, slice):
            # Construct explicit list of indices from slice
            n = np.arange(len(self.samples_indices))[n]

        if hasattr(n, "__iter__"):
            mask = np.full(len(self.data), False)
            samples_indices = np.full((len(n), 2), 0)

            # Create a boolean mask array selecting only array elements we need
            for i, nsample in enumerate(n):
                curi = indices[nsample]
                mask[curi[0]:curi[1]] = True

                # The samples indices must be offset by the number of omitted
                # array elements before their array index
                previous = mask[:curi[0]]
                offset = np.size(previous) - np.count_nonzero(previous)
                samples_indices[i, :] = indices[nsample] - offset

            data = self.data[mask]

            new_instance = self.__class__(
                data, columns = self.columns, **self.attrs
            )
            new_instance.samples_indices = samples_indices
            return new_instance

        # Otherwise return a single sample
        while n < 0:
            n += len(self)

        data = self.data[indices[n, 0]:indices[n, 1]]
        return self.__class__(
            data, len(data), columns = self.columns, **self.attrs
        )


    def __iter__(self):
        # Defined so the class can be iterated as
        # `for sample in class_instance: ...`
        return self


    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration

        self._index += 1
        return self[self._index - 1]




class AsyncIterableSamples(PEPTObject):
    '''Asynchronously apply a function to some samples of data and return those
    processed samples on demand.

    For example, samples of `Cutpoints` are computed from samples of
    `LineData`; `cutpoints_instance[0]` processes the first sample of lines and
    returns it.

    Attributes
    ----------
    samples : instance or subclass of IterableSamples
        The samples of data to be processed; must be a subclass of
        `IterableSamples` to allow iterating over samples (e.g. `LineData`).

    function : callable, signature `func(sample)`
        A function transforming a raw sample from `samples` into a processed
        sample.

    columns : List[str] or None, optional
        The column names of the processed samples.

    executor : concurrent.futures.Executor subclass, default ThreadPoolExecutor
        The executor used

    '''


    def __init__(
        self,
        samples,
        function,
        args = (),
        kwargs = dict(),
        columns = None,
        save_cache = False,
        verbose = True,
    ):
        # Type-checking inputs
        if not isinstance(samples, IterableSamples):
            raise TypeError((
                "The input `samples` must be a collection that allows "
                "iteration in samples - and therefore to be a subclass of "
                "`IterableSamples`"
            ))

        if not callable(function):
            raise TypeError("The input `function` must be callable!")

        # Setting class attributes
        self._samples = samples
        self._function = function
        self._args = tuple(args)
        self._kwargs = dict(kwargs)

        self._columns = None if columns is None else [str(c) for c in columns]

        self._save_cache = bool(save_cache)

        self._index = 0
        self._processed = [None for _ in range(len(self._samples))]


    @property
    def samples(self):
        # The samples of data that will be processed
        return self._samples


    @property
    def columns(self):
        return self._columns


    @property
    def processed(self):
        # Return the list of processed samples
        return self._processed


    @property
    def function(self):
        # The function that transforms `samples` asynchronously
        return self._function


    @property
    def save_cache(self):
        return self._save_cache


    @save_cache.setter
    def save_cache(self, new_save_cache):
        self._save_cache = bool(new_save_cache)


    @property
    def data(self):
        # Accessing all processed data will trigger a full processing run.
        # Defined here to allow e.g. plotting
        return np.vstack(self.traverse(verbose = False))


    def traverse(
        self,
        sample_indices = ...,
        executor = ThreadPoolExecutor,
        max_workers = None,
        verbose = True,
    ):
        '''Apply `self.function` to all samples in `samples` at indices
        `samples_indices`.

        If `save_cache` is `True`, the processed samples are also cached in the
        `data` attribute. Otherwise, they are only returned as a list.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples to process. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all samples).

        verbose : bool, default True
            Show extra information as the processing is done.

        Returns
        -------
        list
            A list of the processed samples, selected by `sample_indices`. The
            type depends on the output of `function`.

        Notes
        -----
        This method is automatically called if the instantiation of the class
        sets `traverse = True`.

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis.
        if sample_indices is Ellipsis:
            sample_indices = np.arange(len(self.samples))
        elif not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        with executor(max_workers = max_workers) as exe:
            # Use pre-computed voxels and voxellise the other samples
            selected_samples = [None for _ in range(len(sample_indices))]
            selected_futures = [None for _ in range(len(sample_indices))]

            # Apply `function` to each selected sample
            for i, n in enumerate(sample_indices):
                # Optimisation: if this sample was already processed and
                # cached, reuse it
                if self.processed[n] is not None:
                    selected_samples[i] = self.processed[n]
                    continue

                # Otherwise, process the sample asynchronously
                selected_futures[i] = exe.submit(
                    self.function,
                    self.samples[n],
                    *self._args,
                    **self._kwargs,
                )

            if verbose:
                sample_indices = tqdm(sample_indices)

            # Iterate through all the futures; if not None (i.e. we processed
            # the just sample now), extract it. Otherwise it was pre-computed
            for i, n in enumerate(sample_indices):
                if selected_futures[i] is not None:
                    selected_samples[i] = selected_futures[i].result()

                    # Delete the future object to release its memory
                    selected_futures[i] = None

                    # If we processed this sample and we use "save_cache",
                    # cache the result in self.voxels
                    if self.save_cache:
                        self.processed[n] = selected_samples[i]

        return selected_samples


    def accumulate(
        self,
        sample_indices = ...,
        op = operator.add,
        executor = ThreadPoolExecutor,
        max_workers = None,
        verbose = True,
    ):
        '''Accumulate all selected processed samples onto the same object using
        the operator `op`.

        For example, this method can be used to voxellise multiple samples of
        lines into the same `Voxels` class. The computation is done in parallel
        and uses the least amount of memory possible.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples to process. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all samples).

        verbose : bool, default True
            Show extra information as the processing runs.

        Returns
        -------
        `type(function(samples[0]))`
            The processed object onto which all processed samples were
            superimposed.

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis.
        if sample_indices is Ellipsis:
            sample_indices = np.arange(len(self.samples))
        elif not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        # Voxellise each selected sample
        with executor(max_workers = max_workers) as exe:
            # Use pre-computed samples and process the other samples
            selected_futures = [None for _ in range(len(sample_indices))]

            for i, n in enumerate(sample_indices):
                # Optimisation: if this sample was already processed, reuse it
                if self._voxels[n] is not None:
                    continue

                # Otherwise, process the sample asynchronously
                selected_futures[i] = exe.submit(
                    self.function,
                    self.samples[n],
                    *self._args,
                    **self._kwargs,
                )

            if verbose:
                sample_indices = tqdm(sample_indices)

            # Iterate through all the futures; if not None (i.e. we voxellised
            # the sample now), get the result. Otherwise it was pre-computed
            if selected_futures[0] is None:
                superimposed = self.processed[0]
            else:
                superimposed = selected_futures[0].result()
                selected_futures[0] = None

            for i in range(1, len(sample_indices)):
                if selected_futures[i] is not None:
                    superimposed = op(
                        superimposed,
                        selected_futures[i].result(),
                    )

                    # Delete future object to release its memory
                    selected_futures[i] = None
                else:
                    superimposed = op(
                        superimposed,
                        self.processed[sample_indices[i]],
                    )

        return superimposed


    def __len__(self):
        return len(self.samples)


    def __getitem__(self, key):
        # For accessing voxels using subscript notation
        key = int(key)

        # Allow negative indices
        while key < 0:
            key += len(self.processed)

        if key >= len(self.processed):
            raise IndexError(textwrap.fill((
                f"The index `{key}` was out of range. There are "
                f"{len(self.samples)} samples to be processed, "
                "indexed from 0."
            )))

        # If the sample was already processed and cached, return it directly
        if self.processed[key] is not None:
            return self.processed[key]

        # Otherwise process it
        p = self.function(
            self.samples[key],      # _memoryview_safe(self.samples[key])
            *self._args,
            **self._kwargs,
        )

        if self.save_cache:
            self.processed[key] = p

        return p


    def __iter__(self):
        # Allow iteration of the class - `for sample in class_instance:`
        return self


    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration

        self._index += 1
        return self[self._index - 1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 21.08.2019


'''Tracer location, identification and tracking algorithms.

The `pept.tracking` subpackage hosts different tracking algorithms, working
with both the base classes, as well as with generic NumPy arrays.

All algorithms here are either ``pept.base.Filter`` or ``pept.base.Reducer``
subclasses, implementing the `.fit` and `.fit_sample` methods; here is an
example using PEPT-ML:

>>> from pept.tracking import *
>>>
>>> cutpoints = Cutpoints(0.5).fit(lines)
>>> clustered = HDBSCAN(0.15).fit(cutpoints)
>>> centres = (SplitLabels() + Centroids() + Stack()).fit(clustered)

Once the processing steps have been tuned (see the `Tutorials`), you can chain
all filters into a `pept.Pipeline` for efficient, parallel execution:

>>> pipeline = (
>>>     Cutpoints(0.5) +
>>>     HDBSCAN(0.15) +
>>>     SplitLabels() + Centroids() + Stack()
>>> )
>>> centres = pipeline.fit(lines)

If you would like to implement a PEPT algorithm, all you need to do is to
subclass a ``pept.base.Filter`` and define the method ``.fit_sample(sample)`` -
and you get parallel execution and pipeline chaining for free!

>>> import pept
>>>
>>> class NewAlgorithm(pept.base.LineDataFilter):
>>>     def __init__(self, setting1, setting2 = None):
>>>         self.setting1 = setting1
>>>         self.setting2 = setting2
>>>
>>>     def fit_sample(self, sample: pept.LineData):
>>>         processed_points = ...
>>>         return pept.PointData(processed_points)

'''


from    .birmingham_method      import  BirminghamMethod
from    .peptml                 import  Cutpoints, Minpoints
from    .peptml                 import  HDBSCAN, HDBSCANClusterer
from    .fpi                    import  FPI

from    .trajectory_separation  import  Segregate

# Import other objects defined elsewhere that should be available here
from    .transformers           import  Stack

from    .transformers           import  SplitLabels
from    .transformers           import  SplitAll

from    .transformers           import  Centroids
from    .transformers           import  LinesCentroids

from    .transformers           import  Condition
from    .transformers           import  Remove

from    .space_transformers     import  Voxelize
from    .space_transformers     import  Interpolate

from    .post                   import  Velocity


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"
